r"""The delete user button and its callback functions."""

# Third party
import streamlit as st

# Local
from streamlit_passwordless import database as db
from streamlit_passwordless.authorization import sign_out
from streamlit_passwordless.bitwarden_passwordless import BitwardenPasswordlessClient
from streamlit_passwordless.models import AdminRole, Role, User

from . import config, core, ids


def _can_delete_user(
    user_id: str, signed_in_user: User | None, allow_delete_self: bool, authorized_for_role: Role
) -> None:
    r"""Check if a user is allowed to delete a user.

    Parameters
    ----------
    user_id : str
        The unique ID of the user to delete.

    signed_in_user : streamlit_passwordless.User or None
        The signed in user performing the delete operation.

    allow_delete_self : bool
        If True `signed_in_user` is allowed to delete itself.

    authorized_for_role : streamlit_passwordless.Role
        The minimum role of `signed_in_user` to be authorized to delete a user.
    """

    if signed_in_user is None or not signed_in_user.is_authenticated:
        error_msg = 'A user must be signed in to be able to delete a user!'
        st.session_state[config.SK_DELETE_USER_VALIDATION_ERRORS] = error_msg
        return

    error_msg = ''
    displayname = n if (n := signed_in_user.displayname) is not None else signed_in_user.username

    if signed_in_user.is_authorized(role=authorized_for_role):
        if signed_in_user.user_id == user_id:
            if allow_delete_self:
                st.session_state[config.SK_DELETE_USER_SELF] = True
            else:
                error_msg = f'Cannot delete signed in user: "{displayname}"!'
        else:
            pass

    else:
        error_msg = f'User "{displayname}" is not authorized to delete users'

    st.session_state[config.SK_DELETE_USER_VALIDATION_ERRORS] = error_msg


def _delete_user(
    client: BitwardenPasswordlessClient, db_session: db.Session, user: db.models.User
) -> tuple[bool, str]:
    r"""Delete a user from Bitwarden Passwordless and the database.

    Parameters
    ----------
    client : streamlit_passwordless.BitwardenPasswordlessClient
        An instance of the Bitwarden Passwordless client to
        communicate with the backend API.

    db_session : streamlit_passwordless.db.Session
        An active session to the Streamlit Passwordless database.

    user : streamlit_passwordless.db.models.User
        The user to delete.

    Returns
    -------
    success : bool
        True if the `user` was successfully deleted and False otherwise.

    error_msg : str
        An error message to display to the user if there was an issue with deleting
        the user. If no error occurred an empty string is returned.
    """

    success, error_msg = core.delete_user_from_bitwarden_passwordless(
        client=client, user_id=user.user_id
    )
    if success:
        success, error_msg = core.delete_user_in_database(session=db_session, user=user)

    return success, error_msg


def delete_user_button(
    client: BitwardenPasswordlessClient,
    db_session: db.Session,
    user: db.models.User | None,
    allow_delete_self: bool = False,
    authorized_for_role: Role = AdminRole,
    label: str = 'Delete User',
    button_type: core.ButtonType = 'primary',
    help: str | None = 'Delete a user from the database and Bitwarden Passwordless.',
    use_container_width: bool = False,
    banner_container: core.BannerContainer | None = None,
    key=ids.DELETE_USER_BUTTON,
) -> tuple[bool, bool]:
    r"""Render the button for deleting a user.

    The user is deleted from Bitwarden Passwordless and the local database.

    Parameters
    ----------
    client : streamlit_passwordless.BitwardenPasswordlessClient
        An instance of the Bitwarden Passwordless client to
        communicate with the backend API.

    db_session : streamlit_passwordless.db.Session
        An active session to the Streamlit Passwordless database.

    user : streamlit_passwordless.db.models.User or None
        The user to delete. If None the delete button is disabled.

    allow_delete_self : bool, default False
        If True the signed in user is allowed to delete itself.

    authorized_for_role : streamlit_passwordless.Role, default streamlit_passwordless.AdminRole
        The minimum role of the signed in user to be authorized to delete a user.

    label : str, default 'Delete User'
        The label of the delete button.

    button_type : Literal['primary', 'secondary'], default 'primary'
        The styling of the delete button. Emulates the `type` parameter of :func:`streamlit.button`.

    help : str or None, default 'Delete a user from the database and Bitwarden Passwordless.'
        The help text to display for the delete button. If None the help text is removed.

    use_container_width : bool, default False
        If True the button will take up the entire width of its enclosing container.

    banner_container : streamlit_passwordless.BannerContainer or None, default None
        A container produced by :func:`streamlit.empty`, in which error or success messages about
        the delete user process will be displayed. Useful to make the banner appear at the desired
        location on a page. If None the banner will be displayed right above the delete button.

    key : str, default streamlit_passwordless.components.ids.DELETE_USER_BUTTON
        The unique identifier of the button. Each component on a page must have a unique key.

    Returns
    -------
    clicked : bool
        True if the button was clicked and False otherwise.

    success : bool
        True if the user deletion was successful and False otherwise.
    """

    banner_container = st.empty() if banner_container is None else banner_container
    signed_in_user = config.get_current_user()

    if user is None:
        disabled = True
        user_id = ''
        username = ''
    else:
        user_id = user.user_id
        username = user.username
        disabled = True if signed_in_user is None else not signed_in_user.is_authenticated

    clicked = st.button(
        label=label,
        type=button_type,
        disabled=disabled,
        on_click=_can_delete_user,
        kwargs={
            'user_id': user_id,
            'signed_in_user': signed_in_user,
            'allow_delete_self': allow_delete_self,
            'authorized_for_role': authorized_for_role,
        },
        use_container_width=use_container_width,
        help=help,
        key=key,
    )

    if not clicked:
        return False, False

    if error_msg := st.session_state.get(config.SK_DELETE_USER_VALIDATION_ERRORS):
        core.display_banner_message(
            message=error_msg,
            message_type=core.BannerMessageType.ERROR,
            container=banner_container,
        )

        return True, False

    if user is None:
        core.display_banner_message(
            message='Cannot delete non-existing user!',
            message_type=core.BannerMessageType.ERROR,
            container=banner_container,
        )

        return True, False

    success, error_msg = _delete_user(client=client, db_session=db_session, user=user)
    if not success:
        core.display_banner_message(
            message=error_msg,
            message_type=core.BannerMessageType.ERROR,
            container=banner_container,
        )

        return True, False

    if st.session_state.get(config.SK_DELETE_USER_SELF, False):
        sign_out(user=signed_in_user)

    core.display_banner_message(
        message=f'Successfully deleted user {username}!',
        message_type=core.BannerMessageType.SUCCESS,
        container=banner_container,
    )

    return True, True



r"""The register-form component and its callback functions."""

# Standard library
import logging
from datetime import timedelta
from typing import Literal

# Third party
import streamlit as st

# Local
from streamlit_passwordless import database as db
from streamlit_passwordless import exceptions, models
from streamlit_passwordless.bitwarden_passwordless.backend import BitwardenPasswordlessClient
from streamlit_passwordless.bitwarden_passwordless.frontend import register_button
from streamlit_passwordless.web import get_origin_header

from . import config, core, ids

logger = logging.getLogger(__name__)


USE_DEFAULT_HELP = '__default__'

USERNAME_HELP = 'A unique identifier for the user.'

DISPLAYNAME_HELP = 'A descriptive name of the user.'

CREDENTIAL_NICKNAME_HELP = (
    'A nickname for the passkey credential to make it easier to '
    'identify which device it belongs to.'
)

DISCOVERABILITY_HELP = (
    'A discoverable passkey allows you to sign in without entering your username '
    'in contrast to a non-discoverable passkey. A non-discoverable does not consume '
    'a passkey slot on a YubiKey, which is the case for a discoverable passkey.'
)

ALIAS_HELP = (
    'One or more aliases that can be used to sign in to the account. '
    'Aliases are separated by semicolon (";"). The username is always '
    'added as an alias. An alias must be unique across all users.'
)


def _render_discoverability_component(
    label: str,
    component_type: Literal['toggle', 'radio'],
    default: bool,
    radio_button_option_names: tuple[str, str],
    radio_button_horizontal: bool,
    help_text: str | None,
    form_type: Literal['new_user', 'existing_user'],
    disabled: bool,
) -> bool:
    r"""Render the discoverability component for the register form.

    The component can be either a toggle switch or a radio button.

    Parameters
    ----------
    label : str
        The label of the discoverability component.

    component_type : Literal['toggle', 'radio']
        If the discoverability component should be rendered as a toggle switch or a radio button.

    default : bool
        The default option of the discoverability component. True means 'discoverable' and
        False 'non-discoverable'.

    radio_button_option_names : tuple[str, str]
        The option names of the radio button if `discoverability_component_type` is set to 'radio'.

    radio_button_horizontal : bool
       The radio button options are is rendered horizontally if True and vertically if False
       if `discoverability_component_type` is set to 'radio'.

    help_text : str or None
        The help text to display for the discoverability component. If '__default__'
        a sensible default help text is used and if None the help text is removed.

    form_type : Literal['new_user', 'existing_user']
        The form (`bitwarden_register_form`, `bitwarden_register_form_existing_user`) in which
        the component is rendered. Used for setting the ID:s of the toggle switch or radio button.
        This is important to avoid the error `streamlit.DuplicateWidgetID` if both
        `bitwarden_register_form` and `bitwarden_register_form_existing_user` are rendered on
        the same page in an application.

    disabled : bool
        True if the component should be disabled and False for enabled.

    Returns
    -------
    discoverable : bool
        The selected option for the discoverability, where True means 'discoverable'
        and False 'non-discoverable'.

    Raises
    ------
    exceptions.StreamlitPasswordlessError
        If invalid options for `component_type` or `form_type` are specified.
    """

    if form_type == 'new_user':
        toggle_switch_id = ids.BP_REGISTER_FORM_DISCOVERABILITY_TOGGLE_SWITCH
        radio_button_id = ids.BP_REGISTER_FORM_DISCOVERABILITY_RADIO_BUTTON
    elif form_type == 'existing_user':
        toggle_switch_id = ids.BP_REGISTER_FORM_EXISTING_USER_DISCOVERABILITY_TOGGLE_SWITCH
        radio_button_id = ids.BP_REGISTER_FORM_EXISTING_USER_DISCOVERABILITY_RADIO_BUTTON
    else:
        raise exceptions.StreamlitPasswordlessError(
            f"Invalid value for {form_type=}. Expected 'new_user' or 'existing_user'."
        )

    _help = DISCOVERABILITY_HELP if help_text == USE_DEFAULT_HELP else help_text
    if component_type == 'toggle':
        discoverable = st.toggle(
            label=label, value=default, help=_help, key=toggle_switch_id, disabled=disabled
        )
    elif component_type == 'radio':
        option_format_mapping = {
            True: radio_button_option_names[0],
            False: radio_button_option_names[1],
        }
        discoverable_radio = st.radio(  # May return none if index option is not specified.
            label=label,
            index=0 if default is True else 1,
            options=(True, False),
            format_func=lambda v: option_format_mapping[v],
            horizontal=radio_button_horizontal,
            help=_help,
            key=radio_button_id,
            disabled=disabled,
        )
        discoverable = True if discoverable_radio is None else discoverable_radio
    else:
        raise exceptions.StreamlitPasswordlessError(
            f"Invalid value for {component_type=}. Expected 'toggle' or 'radio'."
        )

    return discoverable


@st.cache_data(hash_funcs={models.Role: lambda r: r.name})
def _create_user(
    username: str,
    displayname: str | None = None,
    role: models.Role | None = None,
    email: str | None = None,
    aliases: str | None = None,
) -> models.User:
    r"""Create a new user to register.

    Parameters
    ----------
    username : str
        The username.

    displayname : str or None, default None
        The optional displayname of the user.

    role : models.Role or None, default None
        The role to assign to the user. If None the default role of
        :class:`streamlit_passwordless.User` is assigned.

    email : str or None, default None
        The optional email address to associate with the user.

    aliases : str or None, default None
        The optional aliases of the user as a semicolon separated string.

    Returns
    -------
    user : models.User
        The user to create.
    """

    user = models.User(
        username=username.strip(),
        displayname=None if displayname is None else displayname.strip(),
        role=models.UserRole if role is None else role,
        emails=[models.Email(email=email.strip().casefold(), rank=1)] if email else [],
        aliases=aliases,
    )
    logger.debug(f'Successfully created user: {user}')

    return user


@st.cache_data(
    ttl=timedelta(minutes=2),
    show_spinner=False,
    hash_funcs={BitwardenPasswordlessClient: hash, models.User: hash},
)
def _create_register_token(
    client: BitwardenPasswordlessClient, user: models.User, discoverable: bool | None = None
) -> tuple[str, str]:
    r"""Create a register token to register a new passkey with the user's device.

    Parameters
    ----------
    client : BitwardenPasswordlessClient
        The Bitwarden Passwordless client to use for interacting with
        the Bitwarden Passwordless API.

    user : models.User
        The user to register.

    discoverable : bool or None, default None
        If True create a discoverable passkey and if False a non-discoverable passkey.
        If None the setting for discoverability from `client.register_config` is used.

    Returns
    -------
    register_token : str
        The register token. An empty string is returned if an error occurred.

    error_msg : str
        An error message if the register token could not be created.
        An empty string is returned if no error occurred.
    """

    error_msg = ''

    try:
        register_token = client.create_register_token(user=user, discoverable=discoverable)
    except exceptions.RegisterUserError as e:
        logger.error(e.detailed_message)
        error_msg = e.displayable_message
        register_token = ''

    return register_token, error_msg


def _validate_username_field(
    db_session: db.Session, client: BitwardenPasswordlessClient, origin: str, pre_authorized: bool
) -> tuple[bool, bool]:
    r"""Validate the username text input field of the register form.

    Parameters
    ----------
    db_session : streamlit_passwordless.database.Session
        An active database session.

    client : streamlit_passwordless.BitwardenPasswordlessClient
        The Bitwarden Passwordless client to use for interacting with
        the Bitwarden Passwordless backend API.

    origin : str
        The domain name of the application. Used to fetch existing passkey credentials of the
        the user for the application. An authenticated user with existing credentials to the
        application may create new credentials.

    pre_authorized : bool
        If True require a user with the input username to exist in the database to allow
        the user to register a new passkey credential. If False omit this validation.

    Returns
    -------
    bool
        True if the username is valid for registration and False otherwise.

    bool
        True if the user that the supplied username is linked to is already authenticated
        and False otherwise.
    """

    validation_errors = st.session_state[config.SK_REGISTER_FORM_VALIDATION_ERRORS]
    validation_error_field = core.FormField.USERNAME
    username = st.session_state[ids.BP_REGISTER_FORM_USERNAME_TEXT_INPUT]

    if not username:
        validation_errors[validation_error_field] = 'The username field is required!'
        return False, False

    db_user, error_msg = core.get_user_from_database(
        session=db_session, username=username.strip().casefold()
    )
    st.session_state[config.SK_DB_USER] = db_user

    if db_user is None:
        if error_msg:  # DatabaseError
            validation_errors[validation_error_field] = error_msg
            return False, False
        elif not pre_authorized:
            return True, False

        error_msg = (
            f'User {username} does not exist, but is required to exist to allow registration!'
        )
        logger.warning(error_msg)
        validation_errors[validation_error_field] = error_msg
        return False, False

    if (user := config.get_current_user()) is None:
        is_authenticated = False
    else:
        is_authenticated = user.is_authenticated and db_user.user_id == user.user_id

    if is_authenticated:
        return True, True

    if origin:
        try:
            credentials = client.get_credentials(user_id=db_user.user_id, origin=origin)
        except exceptions.StreamlitPasswordlessError as e:
            error_msg = (
                f'Could not get passkey credentials for user {username}!\n{e.detailed_message}'
            )
            logger.error(error_msg)
            credentials = []

        if not credentials:
            return True, False

    error_msg = f'User {username} already exist! To add additional passkeys, please sign in first.'
    validation_errors[validation_error_field] = error_msg

    return False, False


def _validate_email_field(
    db_session: db.Session, email_is_username: bool, username: str | None
) -> bool:
    r"""Validate the email text input field of the register form.

    Parameters
    ----------
    db_session : streamlit_passwordless.database.Session
        An active database session.

    email_is_username : bool
        If True the username is the email address of the user and if False
        the username is distinct from an optional email address of the user.

    username : str or None
        The username of the user to create.

    Returns
    -------
    bool
        True if the email is valid for registration and False otherwise.
    """

    if email_is_username:
        email = username
        validation_error_field = core.FormField.USERNAME
    else:
        email = st.session_state.get(ids.BP_REGISTER_FORM_EMAIL_TEXT_INPUT)
        validation_error_field = core.FormField.EMAIL

    if email:
        validation_errors = st.session_state[config.SK_REGISTER_FORM_VALIDATION_ERRORS]
        db_email, error_msg = core.get_email_from_database(
            session=db_session, email=email.strip().casefold()
        )
        if error_msg:  # DatabaseError
            validation_errors[validation_error_field] = error_msg
            return False
        if db_email is not None:
            error_msg = f'{"User" if email_is_username else "Email"} {email} already exists!'
            validation_errors[validation_error_field] = error_msg
            return False

    return True


def _validate_credential_nickname_field() -> bool:
    r"""Validate the credential nickname text input field of the register form.

    Returns
    -------
    bool
        True if the credential nickname is valid for registration and False otherwise.
    """

    credential_nickname = st.session_state[ids.BP_REGISTER_FORM_CREDENTIAL_NICKNAME_TEXT_INPUT]

    if not credential_nickname.strip():
        error_msg = 'A passkey nickname is required!'
        validation_errors = st.session_state[config.SK_REGISTER_FORM_VALIDATION_ERRORS]
        validation_errors[core.FormField.CREDENTIAL_NICKNAME] = error_msg

        return False

    return True


def _validate_form(
    db_session: db.Session,
    client: BitwardenPasswordlessClient,
    origin: str,
    pre_authorized: bool,
    email_is_username: bool,
    username: str,
    require_credential_nickname: bool,
) -> None:
    r"""Validate the input fields of the register form.

    Updates the following session state keys:
    - `streamlit_passwordless.SK_DB_USER` :
        Assigns the database user object. None is assigned if the user does
        not exist in the database.

    - `streamlit_passwordless.SK_REGISTER_FORM_IS_VALID` :
        True if the form validation passed and False otherwise.

    - `streamlit_passwordless.SK_REGISTER_FORM_VALIDATION_ERRORS` :
        A dictionary mapping of the user form field name to its error message.

    Parameters
    ----------
    db_session : streamlit_passwordless.database.Session
        An active database session.

    client : streamlit_passwordless.BitwardenPasswordlessClient
        The Bitwarden Passwordless client to use for interacting with
        the Bitwarden Passwordless backend API.

    origin : str
        The domain name of the application. Used to fetch existing passkey credentials of the
        the user for the application. An authenticated user with existing credentials to the
        application may create new credentials.

    pre_authorized : bool
        If True require a user with the input username to exist in the database to allow
        the user to register a new passkey credential. If False omit this validation.

    email_is_username : bool
        If True the username is the email address of the user and if False
        the username is distinct from an optional email address of the user.

    username : str
        The username of the user to create.

    require_credential_nickname : bool, default True
        If True the user is required to enter a value for the credential_nickname
        field to be able to register a passkey. If False the validation is omitted.
    """

    username_is_valid, is_authenticated = _validate_username_field(
        db_session=db_session, client=client, origin=origin, pre_authorized=pre_authorized
    )

    if is_authenticated:
        email_is_valid = True
    else:
        email_is_valid = _validate_email_field(
            db_session=db_session, email_is_username=email_is_username, username=username
        )

    if require_credential_nickname:
        credential_nickname_is_valid = _validate_credential_nickname_field()
    else:
        credential_nickname_is_valid = True

    st.session_state[config.SK_REGISTER_FORM_IS_VALID] = (
        username_is_valid and email_is_valid and credential_nickname_is_valid
    )


def bitwarden_register_form(
    client: BitwardenPasswordlessClient,
    db_session: db.Session,
    role: models.Role | None = None,
    pre_authorized: bool = False,
    with_displayname: bool = False,
    with_email: bool = True,
    email_is_username: bool = False,
    with_credential_nickname: bool = True,
    require_credential_nickname: bool = True,
    with_discoverability: bool = False,
    with_alias: bool = False,
    title: str = '#### Register a new passkey with your device',
    border: bool = True,
    validate_button_label: str = 'Validate',
    register_button_label: str = 'Register',
    validate_button_type: core.ButtonType | None = None,
    register_button_type: core.ButtonType = 'primary',
    clear_on_validate: bool = False,
    banner_container: core.BannerContainer | None = None,
    redirect: core.Redirectable | None = None,
    created_by_user_id: str | None = None,
    username_label: str = 'Username',
    username_max_length: int | None = 50,
    username_placeholder: str | None = 'john.doe',
    username_help: str | None = '__default__',
    displayname_label: str = 'Displayname',
    displayname_max_length: int | None = 50,
    displayname_placeholder: str | None = 'John Doe',
    displayname_help: str | None = '__default__',
    email_label: str = 'Email',
    email_max_length: int | None = 50,
    email_placeholder: str | None = 'john.doe@example.com',
    email_help: str | None = 'An email address to associate with the user.',
    credential_nickname_label: str = 'Passkey Nickname',
    credential_nickname_max_length: int | None = 50,
    credential_nickname_placeholder: str | None = 'Bitwarden or YubiKey-5C-NFC',
    credential_nickname_help: str | None = '__default__',
    discoverability_label: str = 'Discoverable Passkey',
    discoverability_component_type: Literal['toggle', 'radio'] = 'toggle',
    discoverability_default_option: bool = True,
    discoverability_radio_button_options: tuple[str, str] = ('Discoverable', 'Non-Discoverable'),
    discoverability_radio_button_horizontal: bool = True,
    discoverability_help: str | None = '__default__',
    alias_label: str = 'Alias',
    alias_max_length: int | None = 50,
    alias_placeholder: str | None = 'j;john;jd',
    alias_help: str | None = '__default__',
) -> tuple[models.User | None, bool]:
    r"""Render the Bitwarden Passwordless register form.

    Allows the user to register an account with the application by creating
    and registrering a passkey with the user's device.

    Parameters
    ----------
    client : streamlit_passwordless.BitwardenPasswordlessClient
        The Bitwarden Passwordless client to use for interacting with
        the Bitwarden Passwordless application.

    db_session : streamlit_passwordless.db.Session
        An active database session.

    role : streamlit_passwordless.Role or None, default None
        The role to assign to the user. If None the user will be assigned the default
        user role of :class:`streamlit_passwordless.User` i.e. a role with name
        :attr:`streamlit_passwordless.UserRoleName.USER`.

    pre_authorized : bool, default False
        If True require a user with the input username to exist in the database to allow
        the user to register a new passkey credential. If False omit this validation.

    with_displayname : bool, default False
        If True the displayname field will be added to the form allowing
        the user to fill out a displayname for the account.

    with_email : bool, default True
        If True the email field is added to the form to enable to
        enter an email address to associate with the user.

    email_is_username : bool, default False
        If True the username field will prompt to enter an email address, which will be used
        as the username and be added as an email address of the user.

    with_credential_nickname : bool, default True
        If True the credential_nickname field will be added to the form allowing the user to
        specify a nickname for the passkey credential to create e.g. "YubiKey-5C-NFC".
        If False the username will be used as the `credential_nickname`.

    require_credential_nickname : bool, default True
        If True the user is required to enter a value for the credential_nickname
        field to be able to register a passkey. If False the validation is omitted.

    with_discoverability : bool, default False
        If True the discoverability component is added to the form allowing the user to toggle
        between creating a *discoverable* or *non-discoverable* passkey. A discoverable passkey
        allows the user to sign in without entering the username in contrast to a non-discoverable
        passkey, which is traditionally used for a MFA setup. A discoverable passkey consumes a
        passkey slot on a YubiKey, while a non-discoverable does not. In most cases you want to
        create a discoverable passkey, which is also the default when the component is not enabled.
        The component can be rendered as a toggle switch (default) or a radio button by specifying
        the option `discoverability_component_type`.

    with_alias : bool, default False
        If True the alias field will be added to the form allowing the user to fill
        out one or multiple aliases for the account. Aliases should be separated by
        by semicolon (";"). An alias must be unique across all users.

    title : str, default '#### Register a new passkey with your device'
        The title of the registration form. Markdown is supported.

    border : bool, default True
        If True a border will be rendered around the form.

    validate_button_label : str, default 'Validate'
        The label of the validate button. The validate button validates the fields of the form.

    register_button_label : str, default 'Register'
        The label of the register button. The register button registers a new passkey with the
        user's device. It is enabled when the form is valid.

    validate_button_type : Literal['primary', 'secondary'] or None, default None
        The styling of the validate button. Emulates the `type` parameter of :func:`streamlit.button`.
        If None the button type will be 'secondary' when the form is valid and 'primary' when invalid.

    register_button_type : Literal['primary', 'secondary'], default 'primary'
        The styling of the register button. Emulates the `type` parameter of :func:`streamlit.button`.

    clear_on_validate : bool, default False
        True if the form fields should be cleared when the validate form button is clicked
        and False otherwise.

    banner_container : streamlit_passwordless.BannerContainer or None, default None
        A container produced by :func:`streamlit.empty`, in which error or success messages about
        the register user process will be displayed. Useful to make the banner appear at the desired
        location on a page. If None the banner will be displayed right above the form.

    redirect : str or streamlit.navigation.page.StreamlitPage or None, default None
        The Streamlit page to redirect `user` to on successful registration. If str it should
        be the path, relative to the file passed to the ``streamlit run`` command, to the
        Python file containing the page to redirect to. See :func:`streamlit.switch_page`
        for more info. If None no redirect is performed.

    created_by_user_id : str or None, default None
        The ID of the user that is creating the new user. If None the ID of
        the created user will be used if the user does not already exist.

    Other Parameters
    ----------------
    username_label : str, default 'Username'
        The label of the username field.

    username_max_length : int or None, default 50
        The maximum allowed number of characters of the username field.
        If None the upper limit is removed.

    username_placeholder : str or None, default 'john.doe@example.com'
        The placeholder of the username field. If None the placeholder is removed.

    username_help : str or None, default '__default__'
        The help text to display for the username field. If '__default__' a sensible default
        help text will be used and if None the help text is removed.

    displayname_label : str, default 'Displayname'
        The label of the displayname field.

    displayname_max_length : int or None, default 50
        The maximum allowed number of characters of the displayname field.
        If None the upper limit is removed.

    displayname_placeholder : str or None, default 'John Doe'
        The placeholder of the displayname field. If None the placeholder is removed.

    displayname_help : str or None, default '__default__'
        The help text to display for the displayname field. If '__default__' a sensible default
        help text will be used and if None the help text is removed.

    email_label : str, default 'Email'
        The label of the email field.

    email_max_length : int or None, default 50
        The maximum allowed number of characters of the email field.
        If None the upper limit is removed.

    email_placeholder : str or None, default 'john.doe@example.com'
        The placeholder of the email field. If None the placeholder is removed.

    email_help : str or None, default 'An email address to associate with the user.'
        The help text to display for the email field. If None the help text is removed.

    credential_nickname_label : str, default 'Passkey Nickname'
        The label of the credential_nickname field.

    credential_nickname_max_length : int or None, default 50
        The maximum allowed number of characters of the credential_nickname field.
        If None the upper limit is removed.

    credential_nickname_placeholder : str or None, default 'Bitwarden or YubiKey-5C-NFC'
        The placeholder of the credential_nickname field. If None the placeholder is removed.

    credential_nickname_help : str or None, default '__default__'
        The help text to display for the credential_nickname field. If '__default__' a sensible
        default help text is used and if None the help text is removed.

    discoverability_label : str, default 'Discoverable Passkey'
        The label of the discoverability component.

    discoverability_component_type : Literal['toggle', 'radio'], default 'toggle'
        If the discoverability component should be rendered as a toggle switch or a radio button.

    discoverability_default_option : bool, default True
        The default option of the discoverability component. True means 'discoverable' and
        False 'non-discoverable'.

    discoverability_radio_button_options : tuple[str, str], default ('Discoverable', 'Non-Discoverable')
        The option names of the radio button if `discoverability_component_type` is set to 'radio'.

    discoverability_radio_button_horizontal : bool, default True
       The radio button options are rendered horizontally if True and vertically if False
       if `discoverability_component_type` is set to 'radio'.

    discoverability_help : str or None, default '__default__'
        The help text to display for the discoverability component. If '__default__' a sensible
        default help text is used and if None the help text is removed.

    alias_label : str, default 'Alias'
        The label of the alias field.

    alias_max_length : int or None, default 50
        The maximum allowed number of characters of the alias field.
        If None the upper limit is removed.

    alias_placeholder : str or None, default 'j;john;jd'
        The placeholder of the alias field. If None the placeholder is removed.

    alias_help : str or None, default '__default__'
        The help text to display for the alias field. If '__default__' a sensible default
        help text will be used and if None the help text is removed.

    Returns
    -------
    user : streamlit_passwordless.User or None
        The user object of the user that registered a passkey credential. None is returned if a user
        has not registered yet or if the registration failed and a user object could not be retrieved.

    bool
        True if the created user has registered a passkey and False otherwise.
    """

    user = None
    error_msg = ''
    banner_container = st.empty() if banner_container is None else banner_container
    banner_container_mapping = {}

    with st.container(border=border):
        st.markdown(title)
        with st.form(key=ids.BP_REGISTER_FORM, clear_on_submit=clear_on_validate, border=False):
            if email_is_username:
                username_label = email_label
                username_placeholder = email_placeholder
                username_max_length = email_max_length

            banner_container_mapping[core.FormField.USERNAME] = st.empty()
            username = st.text_input(
                label=username_label,
                placeholder=username_placeholder,
                max_chars=username_max_length,
                help=USERNAME_HELP if username_help == USE_DEFAULT_HELP else username_help,
                key=ids.BP_REGISTER_FORM_USERNAME_TEXT_INPUT,
            )

            if with_displayname:
                displayname = st.text_input(
                    label=displayname_label,
                    placeholder=displayname_placeholder,
                    max_chars=displayname_max_length,
                    help=(
                        DISPLAYNAME_HELP
                        if displayname_help == USE_DEFAULT_HELP
                        else displayname_help
                    ),
                    key=ids.BP_REGISTER_FORM_DISPLAYNAME_TEXT_INPUT,
                )
            else:
                displayname = None

            if with_email:
                if not email_is_username:
                    banner_container_mapping[core.FormField.EMAIL] = st.empty()
                    email = st.text_input(
                        label=email_label,
                        placeholder=email_placeholder,
                        max_chars=email_max_length,
                        help=email_help,
                        key=ids.BP_REGISTER_FORM_EMAIL_TEXT_INPUT,
                    )
                else:
                    email = username
            else:
                email = None

            if with_credential_nickname:
                banner_container_mapping[core.FormField.CREDENTIAL_NICKNAME] = st.empty()
                credential_nickname = st.text_input(
                    label=credential_nickname_label,
                    placeholder=credential_nickname_placeholder,
                    max_chars=credential_nickname_max_length,
                    help=(
                        CREDENTIAL_NICKNAME_HELP
                        if credential_nickname_help == USE_DEFAULT_HELP
                        else credential_nickname_help
                    ),
                    key=ids.BP_REGISTER_FORM_CREDENTIAL_NICKNAME_TEXT_INPUT,
                )
                credential_nickname = credential_nickname.strip()
            else:
                credential_nickname = username

            if with_discoverability:
                discoverable = _render_discoverability_component(
                    label=discoverability_label,
                    component_type=discoverability_component_type,
                    default=discoverability_default_option,
                    radio_button_option_names=discoverability_radio_button_options,
                    radio_button_horizontal=discoverability_radio_button_horizontal,
                    help_text=discoverability_help,
                    form_type='new_user',
                    disabled=False,
                )
            else:
                discoverable = None

            if with_alias:
                aliases = st.text_input(
                    label=alias_label,
                    placeholder=alias_placeholder,
                    max_chars=alias_max_length,
                    help=ALIAS_HELP if alias_help == USE_DEFAULT_HELP else alias_help,
                    key=ids.BP_REGISTER_FORM_ALIASES_TEXT_INPUT,
                )
            else:
                aliases = None

            form_is_valid = st.session_state[config.SK_REGISTER_FORM_IS_VALID]
            if validate_button_type is None:
                button_type: core.ButtonType = 'secondary' if form_is_valid else 'primary'
            else:
                button_type = validate_button_type

            try:
                origin = get_origin_header()
            except exceptions.StreamlitPasswordlessError as e:
                logger.error(e.detailed_message)
                error_msg = e.displayable_message
                origin = ''

            st.form_submit_button(
                label=validate_button_label,
                type=button_type,
                on_click=_validate_form,
                kwargs={
                    'db_session': db_session,
                    'client': client,
                    'origin': origin,
                    'pre_authorized': pre_authorized,
                    'email_is_username': email_is_username,
                    'username': username,
                    'require_credential_nickname': require_credential_nickname,
                },
            )

        db_user = st.session_state.get(config.SK_DB_USER)
        register_token = ''

        if form_is_valid:
            if db_user:
                user = models.User.model_validate(db_user)
            else:
                user = _create_user(
                    username=username,
                    displayname=displayname,
                    role=role,
                    email=email,
                    aliases=aliases,
                )
            register_token, error_msg = _create_register_token(
                client=client, user=user, discoverable=discoverable
            )

        form_is_not_valid = not form_is_valid

        token, error, clicked = register_button(
            register_token=register_token,
            public_key=client.public_key,
            credential_nickname=credential_nickname,
            disabled=form_is_not_valid,
            label=register_button_label,
            button_type=register_button_type,
            key=ids.BP_REGISTER_FORM_SUBMIT_BUTTON,
        )

    if form_is_not_valid:
        validation_error_key = config.SK_REGISTER_FORM_VALIDATION_ERRORS
        core.process_form_validation_errors(
            validation_errors=st.session_state[validation_error_key],
            banner_container_mapping=banner_container_mapping,  # type: ignore
            default_banner_container=banner_container,
        )
        st.session_state[validation_error_key] = {}

        return user, False

    if not clicked:
        return user, False

    if user is None:
        core.display_banner_message(
            message=error_msg, message_type=core.BannerMessageType.ERROR, container=banner_container
        )
        return user, False

    if not token:
        if error:
            error_msg = f'Error creating passkey for user "{username}"!\nerror : {error}'
        else:
            error_msg = 'Unexpected error for missing register token!'

        logger.error(error_msg)
        core.display_banner_message(
            message=error_msg, message_type=core.BannerMessageType.ERROR, container=banner_container
        )
        return user, False

    final_error_msg = ''
    if not db_user:
        can_save_sign_in_to_db, error_msg = core.create_user_in_database(
            session=db_session,
            user=user,
            created_by_user_id=created_by_user_id if created_by_user_id else user.user_id,
        )
        if not can_save_sign_in_to_db:
            final_error_msg = (
                f'Could not save user {user.username} to database! '
                'A mismatch between Bitwarden Passwordless.dev and local database has occurred!'
            )
    else:
        can_save_sign_in_to_db = True

    # The user is still registered even though the sign in may fail!
    user_sign_in, _ = core.verify_sign_in(client=client, token=token)
    user.sign_in = user_sign_in
    st.session_state[config.SK_USER] = user

    sign_in_failed_error_msg = (
        f'User {username} was registered, but the sign in attempt '
        f'with registered passkey "{credential_nickname}" failed!'
    )
    if user_sign_in is None:
        final_error_msg = f'{final_error_msg}\n{sign_in_failed_error_msg}'
        save_user_sign_in_to_db_ok = False

    elif can_save_sign_in_to_db:
        if not user_sign_in.success:
            final_error_msg = f'{final_error_msg}\n{sign_in_failed_error_msg}'

        save_user_sign_in_to_db_ok, error_msg = core.save_user_sign_in_to_database(
            session=db_session, user_sign_in=user_sign_in
        )
        if not save_user_sign_in_to_db_ok:
            final_error_msg = f'{final_error_msg}\n{error_msg}'
    else:
        save_user_sign_in_to_db_ok = False

    if not can_save_sign_in_to_db:
        logger.error(final_error_msg)
        core.display_banner_message(
            message=final_error_msg,
            message_type=core.BannerMessageType.ERROR,
            container=banner_container,
        )

    elif save_user_sign_in_to_db_ok:
        msg = f'Successfully registered user: {user.username}!'
        logger.info(msg)
        core.display_banner_message(
            message=msg, message_type=core.BannerMessageType.SUCCESS, container=banner_container
        )
        if redirect:
            st.switch_page(redirect)

    elif not save_user_sign_in_to_db_ok:
        logger.warning(final_error_msg)
        core.display_banner_message(
            message=final_error_msg,
            message_type=core.BannerMessageType.WARNING,
            container=banner_container,
        )

    else:
        pass

    return user, True


def bitwarden_register_form_existing_user(
    client: BitwardenPasswordlessClient,
    db_session: db.Session,
    user: models.User | db.models.User | None = None,
    get_current_user: bool = True,
    with_credential_nickname: bool = True,
    require_credential_nickname: bool = True,
    with_discoverability: bool = False,
    title: str = '#### Register a new passkey with your device',
    border: bool = True,
    register_button_label: str = 'Register',
    register_button_type: core.ButtonType = 'primary',
    banner_container: core.BannerContainer | None = None,
    redirect: core.Redirectable | None = None,
    credential_nickname_label: str = 'Passkey Nickname',
    credential_nickname_max_length: int | None = 50,
    credential_nickname_placeholder: str | None = 'Bitwarden or YubiKey-5C-NFC',
    credential_nickname_help: str | None = '__default__',
    discoverability_label: str = 'Discoverable Passkey',
    discoverability_component_type: Literal['toggle', 'radio'] = 'toggle',
    discoverability_default_option: bool = True,
    discoverability_radio_button_options: tuple[str, str] = ('Discoverable', 'Non-Discoverable'),
    discoverability_radio_button_horizontal: bool = True,
    discoverability_help: str | None = '__default__',
) -> tuple[models.User | None, bool]:
    r"""Render the Bitwarden Passwordless register form for an existing user.

    Allows an existing user to register a new passkey credential with the application.
    The register button is disabled if the user is not signed in.

    Parameters
    ----------
    client : streamlit_passwordless.BitwardenPasswordlessClient
        The Bitwarden Passwordless client to use for interacting with
        the Bitwarden Passwordless backend API.

    db_session : streamlit_passwordless.db.Session
        An active database session.

    user : streamlit_passwordless.User or streamlit_passwordless.db.User or None, default None
        The user for which to register a passkey credential. If None the user from the session
        state `streamlit_passwordless.SK_USER` is used (if `get_current_user` is True). If `user`
        is not signed in the register button is disabled. `user.sign_in` is modified with the sign
        in entry with the newly registered passkey after it has been created. The session state
        :attr:`streamlit_passwordless.SK_USER` is updated with the modified `user`. An instance of
        :class:`streamlit_passwordless.db.User` should be supplied for the use case when an admin
        needs to register a passkey for a selected user.

    get_current_user : bool, default True
        If True the current user will be retrieved from the session state if `user` is None.
        If set to False it will ensure that the form is disabled if `user` is None.

    with_credential_nickname : bool, default True
        If True the credential_nickname field will be added to the form allowing the user to
        specify a nickname for the passkey credential to create e.g. "YubiKey-5C-NFC".
        If False the username will be used as the `credential_nickname`.

    require_credential_nickname : bool, default True
        If True the user is required to enter a value for the credential_nickname
        field to be able to register a passkey. If False the validation is omitted.

    with_discoverability : bool, default False
        If True the discoverability component is added to the form allowing the user to toggle
        between creating a *discoverable* or *non-discoverable* passkey. A discoverable passkey
        allows the user to sign in without entering the username in contrast to a non-discoverable
        passkey, which is traditionally used for a MFA setup. A discoverable passkey consumes a
        passkey slot on a YubiKey, while a non-discoverable does not. In most cases you want to
        create a discoverable passkey, which is also the default when the component is not enabled.
        The component can be rendered as a toggle switch (default) or a radio button by specifying
        the option `discoverability_component_type`.

    title : str, default '#### Register a new passkey with your device'
        The title of the registration form. Markdown is supported.

    border : bool, default True
        If True a border will be rendered around the form.

    register_button_label : str, default 'Register'
        The label of the register button to start the passkey registration with the
        user's device. The button is enabled if `user` is signed in and disabled otherwise.

    register_button_type : Literal['primary', 'secondary'], default 'primary'
        The styling of the register button. Emulates the `type` parameter of :func:`streamlit.button`.

    banner_container : streamlit_passwordless.BannerContainer or None, default None
        A container produced by :func:`streamlit.empty`, in which error or success messages about
        the register user process will be displayed. Useful to make the banner appear at the desired
        location on a page. If None the banner will be displayed right above the form.

    redirect : str or streamlit.navigation.page.StreamlitPage or None, default None
        The Streamlit page to redirect `user` to on successful registration. If str it should
        be the path, relative to the file passed to the ``streamlit run`` command, to the
        Python file containing the page to redirect to. See :func:`streamlit.switch_page`
        for more info. If None no redirect is performed.

    Other Parameters
    ----------------
    credential_nickname_label : str, default 'Passkey Nickname'
        The label of the credential_nickname field.

    credential_nickname_max_length : int or None, default 50
        The maximum allowed number of characters of the credential_nickname field.
        If None the upper limit is removed.

    credential_nickname_placeholder : str or None, default 'Bitwarden or YubiKey-5C-NFC'
        The placeholder of the credential_nickname field. If None the placeholder is removed.

    credential_nickname_help : str or None, default '__default__'
        The help text to display for the credential_nickname field. If '__default__' a sensible
        default help text is used and if None the help text is removed.

    discoverability_label : str, default 'Discoverable Passkey'
        The label of the discoverability component.

    discoverability_component_type : Literal['toggle', 'radio'], default 'toggle'
        If the discoverability component should be rendered as a toggle switch or a radio button.

    discoverability_default_option : bool, default True
        The default option of the discoverability component. True means 'discoverable' and
        False 'non-discoverable'.

    discoverability_radio_button_options : tuple[str, str], default ('Discoverable', 'Non-Discoverable')
        The option names of the radio button if `discoverability_component_type` is set to 'radio'.

    discoverability_radio_button_horizontal : bool, default True
       The radio button options are rendered horizontally if True and vertically if False
       if `discoverability_component_type` is set to 'radio'.

    discoverability_help : str or None, default '__default__'
        The help text to display for the discoverability component. If '__default__' a sensible
        default help text is used and if None the help text is removed.

    Returns
    -------
    user_sign_in : streamlit_passwordless.UserSignIn or None
        The `user` modified with the `user.sign_in` entry with the newly registered passkey
        credential. None is returned if the user has not registered a passkey yet or if the
        registration failed.

    bool
        True if `user` has registered a passkey and False otherwise.
    """

    error_msg = ''
    banner_container = st.empty() if banner_container is None else banner_container
    if get_current_user:
        user = st.session_state.get(config.SK_USER) if user is None else user
    username, components_disabled = ('', True) if user is None else (user.username, False)

    with st.container(border=border):
        st.markdown(title)
        if with_credential_nickname:
            credential_nickname = st.text_input(
                label=credential_nickname_label,
                placeholder=credential_nickname_placeholder,
                max_chars=credential_nickname_max_length,
                help=(
                    CREDENTIAL_NICKNAME_HELP
                    if credential_nickname_help == USE_DEFAULT_HELP
                    else credential_nickname_help
                ),
                disabled=components_disabled,
                key=ids.BP_REGISTER_FORM_EXISTING_USER_CREDENTIAL_NICKNAME_TEXT_INPUT,
            )
            credential_nickname = credential_nickname.strip()
        else:
            credential_nickname = username

        if with_discoverability:
            discoverable = _render_discoverability_component(
                label=discoverability_label,
                component_type=discoverability_component_type,
                default=discoverability_default_option,
                radio_button_option_names=discoverability_radio_button_options,
                radio_button_horizontal=discoverability_radio_button_horizontal,
                help_text=discoverability_help,
                form_type='existing_user',
                disabled=components_disabled,
            )
        else:
            discoverable = None

        if isinstance(user, db.models.User):
            from_db_user = True
            is_authenticated = True
            user = models.User.model_validate(user)
        else:
            from_db_user = False
            is_authenticated = False if user is None else user.is_authenticated

        if user is not None and is_authenticated:
            disabled = False
            register_token, error_msg = _create_register_token(
                client=client, user=user, discoverable=discoverable
            )
        else:
            disabled = True
            register_token = ''

        if require_credential_nickname and not credential_nickname:
            disabled = True

        token, error, clicked = register_button(
            register_token=register_token,
            public_key=client.public_key,
            credential_nickname=credential_nickname,
            disabled=disabled,
            label=register_button_label,
            button_type=register_button_type,
            key=ids.BP_REGISTER_FORM_EXISTING_USER_SUBMIT_BUTTON,
        )

    if disabled or not clicked or user is None:
        return user, False

    if not token:
        if error:
            error_msg = f'Error creating passkey for user "{username}"!\nerror : {error}'
        else:
            error_msg = 'Unexpected error for missing register token!'

        logger.error(error_msg)
        core.display_banner_message(
            message=error_msg, message_type=core.BannerMessageType.ERROR, container=banner_container
        )
        return user, False

    # The passkey is still registered even though the sign in may fail!
    user_sign_in, _ = core.verify_sign_in(client=client, token=token)
    user.sign_in = user_sign_in
    if not from_db_user:
        st.session_state[config.SK_USER] = user

    final_error_msg = ''
    sign_in_failed_error_msg = (
        f'Passkey "{credential_nickname}" was registered, '
        'but the sign in attempt with registered passkey failed!'
    )
    if user_sign_in is None:
        final_error_msg = sign_in_failed_error_msg
        save_user_sign_in_to_db_ok = False
    else:
        if not user_sign_in.success:
            final_error_msg = sign_in_failed_error_msg

        save_user_sign_in_to_db_ok, error_msg = core.save_user_sign_in_to_database(
            session=db_session, user_sign_in=user_sign_in
        )
        if not save_user_sign_in_to_db_ok:
            final_error_msg = f'{final_error_msg}\n{error_msg}'

    if save_user_sign_in_to_db_ok:
        msg = f'Successfully registered passkey "{credential_nickname}" for user "{username}"!'
        logger.info(msg)
        core.display_banner_message(
            message=msg, message_type=core.BannerMessageType.SUCCESS, container=banner_container
        )
        if redirect:
            st.switch_page(redirect)
    else:
        logger.warning(final_error_msg)
        core.display_banner_message(
            message=final_error_msg,
            message_type=core.BannerMessageType.WARNING,
            container=banner_container,
        )

    return user, True



r"""The database functions that are cached by Streamlit."""

# Standard library
import logging

# Third party
import streamlit as st

# Local
from streamlit_passwordless import exceptions

from . import core
from .crud.role import create_default_roles

logger = logging.getLogger(__name__)

create_session_factory_cached = st.cache_resource(core.create_session_factory)


@st.cache_resource()
def init(_session: core.Session) -> tuple[bool, str]:
    r"""Initialize a database with the required data.

    The default roles of Streamlit Passwordless are created in the database.

    Parameters
    ----------
    _session : streamlit_passwordless.db.Session
        An active database session.

    Returns
    -------
    error : bool
        True if a :exc:`streamlit_passwordless.DatabaseError` occurred and the database
        could not be initialized correctly and False for no error.

    error_msg : str
        An error message that is safe to display to the user. An empty
        string is returned if `error` is False.
    """

    error = False
    error_msg = ''
    try:
        create_default_roles(session=_session, commit=True)
    except exceptions.DatabaseError as e:
        error = True
        error_msg = e.displayable_message
        logger.warning(e.detailed_message)
        _session.rollback()

    return error, error_msg



r"""The entry point of the sub-command run.

Run the Streamlit Passwordless web apps.
"""

# Standard library
import subprocess
from pathlib import Path
from typing import Iterable

# Third party
import click

# Local
from streamlit_passwordless.app.main import APP_PATH
from streamlit_passwordless.app.pages.init import INIT_PATH

streamlit_args_argument = click.argument('streamlit_args', nargs=-1, type=click.UNPROCESSED)


def run_streamlit_app(path: Path | str, streamlit_args: Iterable[str]) -> None:
    r"""Run a Streamlit application.

    Parameters
    ----------
    path : Path or str
        The path to the module with the Streamlit application to run.

    streamlit_args : tuple[str, ...]
        Additional arguments to pass along to Streamlit.
    """

    run_cmd = ['python', '-m', 'streamlit', 'run', str(path) if isinstance(path, Path) else path]
    run_cmd.extend(streamlit_args)

    click.echo('Launching Streamlit ...')
    subprocess.run(run_cmd)


@click.group()
def run() -> None:
    """Run the Streamlit apps of Streamlit Passwordless.

    \b
    Configuring Streamlit
    ---------------------
    Streamlit can be configured with config files, environment variables or command line options.
    Below is a list of the configuration options in order in which they will override each
    other:

    1 : Command line options.

    2 : Environment variables with names of the config options prefixed by STREAMLIT, e.g. STREAMLIT_SERVER_PORT

    3 : A local config file located in the current working directory at "./.streamlit/config.toml".

    4 : A global config file located in the user's home directory at: "~/.streamlit/config.toml"

    See also the Streamlit documentation on configuration for more details:
    https://docs.streamlit.io/library/advanced-features/configuration#view-all-configuration-options

    \b
    Examples
    --------
    Initialize the Streamlit Passwordless database and pass on command line options to Streamlit:
        $ stp run init --theme.base dark --server.headless true
    """


@run.command(context_settings={'ignore_unknown_options': True})
@streamlit_args_argument
def init(streamlit_args: tuple[str, ...]) -> None:
    """Initialize the Streamlit Passwordless user database."""

    run_streamlit_app(path=INIT_PATH, streamlit_args=streamlit_args)


@run.command(context_settings={'ignore_unknown_options': True})
@streamlit_args_argument
def admin(streamlit_args: tuple[str, ...]) -> None:
    """Run the Streamlit Passwordless admin web app."""

    run_streamlit_app(path=APP_PATH, streamlit_args=streamlit_args)



r"""The schemas for the UserSignIn model."""

# Standard library
from datetime import datetime

# Third party
from pydantic import AnyHttpUrl, field_validator

# Local
from .core import SchemaBaseModel


class UserSignInCreate(SchemaBaseModel):
    r"""The schema for creating a new user sign in entry.

    Parameters
    ----------
    user_id : str
        The unique ID of the user that signed in to the application.

    sign_in_timestamp : datetime
        The timestamp in timezone UTC when the user signed in.

    success : bool
        True if the user was successfully signed in and False otherwise.

    origin : str
        The domain name or IP-address of the application the user signed in to.

    device : str
        The device the user signed in from. E.g. a web browser.

    country : str
        The country code of the country that the user signed in from. E.g. SE for Sweden.

    credential_nickname : str
        The nickname of the passkey that was used to sign in.

    credential_id : str
        The ID of the passkey credential used to sign in.

    sign_in_type : str
        The type of sign in method. E.g. 'passkey_signin'.

    rp_id : str or None
        The ID of the relaying party, which is the server that
        verifies the credentials during the sign in process.
    """

    user_id: str
    sign_in_timestamp: datetime
    success: bool
    origin: str | AnyHttpUrl
    device: str
    country: str
    credential_nickname: str
    credential_id: str
    sign_in_type: str
    rp_id: str | None

    @field_validator('origin')
    def convert_origin_to_string(cls, v: str | AnyHttpUrl) -> str:
        return v if isinstance(v, str) else str(v)



r"""The views of the init page."""

# Third party
import streamlit as st

# Local
import streamlit_passwordless.database as db
from streamlit_passwordless.bitwarden_passwordless import BitwardenPasswordlessClient
from streamlit_passwordless.components import (
    bitwarden_register_form,
    bitwarden_register_form_existing_user,
    bitwarden_sign_in_form,
)
from streamlit_passwordless.components.config import ICON_SUCCESS, ICON_WARNING
from streamlit_passwordless.models import Role


def title() -> None:
    r"""Render the title view of the init page."""

    st.title('Initialize Streamlit Passwordless')
    st.divider()


def initialize(db_session: db.Session, db_url: db.URL) -> None:
    r"""Render the initialize view of the init page.

    Initialize the database by creating the user roles and
    warn the user if the database is already initialized.

    Parameters
    ----------
    db_session : streamlit_passwordless.db.Session
        An active session to the database to initialize.

    db_url : streamlit_passwordless.db.URL
        The database url of the database being initialized.
    """

    error, error_msg = db.init(_session=db_session)
    if error:
        message = f'Database "{db_url}" already initialized! {error_msg}'
        st.warning(message, icon=ICON_WARNING)
    else:
        st.success(f'Successfully initialized database : "{db_url}"!', icon=ICON_SUCCESS)

    st.divider()


def create_admin_user(db_session: db.Session, client: BitwardenPasswordlessClient) -> None:
    r"""Render the create admin user view of the init page.

    Parameters
    ----------
    db_session : streamlit_passwordless.db.Session
        An active session to the database to initialize.

    client : streamlit_passwordless.BitwardenPasswordlessClient
        An instance of the Bitwarden Passwordless client to
        communicate with the Bitwarden Passwordless backend API.
    """

    left_col, right_col = st.columns(2)
    with left_col:
        user, _ = bitwarden_register_form(
            client=client,
            db_session=db_session,
            role=Role.create_admin(),
            with_displayname=True,
            with_credential_nickname=True,
            with_discoverability=True,
            title='Register an admin user',
        )

    username = user.username if user is not None else ''
    with right_col:
        bitwarden_sign_in_form(client=client, db_session=db_session, with_alias=False)
        bitwarden_register_form_existing_user(
            client=client,
            db_session=db_session,
            with_discoverability=True,
            title=f'Register additional passkeys for user : {username}',
        )



r"""Contains the Streamlit configuration of the web components."""

# Standard library

# Third party
import streamlit as st

# Local
from streamlit_passwordless.models import User

# =====================================================================================
# Session state keys
# =====================================================================================

SK_USER = 'stp-user'
SK_USER_SIGN_IN = 'stp-user-sign-in'
SK_DB_USER = 'stp-db-user'
SK_REGISTER_FORM_IS_VALID = 'stp-register-form-is-valid'
SK_REGISTER_FORM_VALIDATION_ERRORS = 'stp-register-form-validation-errors'
SK_ROLES = 'stp-roles'
SK_CUSTOM_ROLES = 'stp-custom-roles'
SK_SESSION_STATE_INITIALIZED = 'stp-session-state-initialized'
SK_CREATE_USER_FORM_IS_VALID = 'stp-create-user-form-is-valid'
SK_CREATE_USER_FORM_VALIDATION_ERRORS = 'stp-create-user-form-validation-errors'
SK_DELETE_USER_VALIDATION_ERRORS = 'stp-delete-user-validation-errors'
SK_DELETE_USER_SELF = 'stp-delete-user-self'
SK_UPDATE_USER_FORM_IS_VALID = 'stp-update-user-form-is-valid'
SK_UPDATE_USER_FORM_VALIDATION_ERRORS = 'stp-update-user-form-validation-errors'

# =====================================================================================
# Icons
# =====================================================================================

ICON_ERROR = '🚨'
ICON_INFO = 'ℹ️'
ICON_SUCCESS = '✅'
ICON_WARNING = '⚠️'


# =====================================================================================
# Functions
# =====================================================================================


def init_session_state() -> None:
    r"""Initialize the session state.

    Session state keys
    ------------------
    SK_SESSION_STATE_INITIALIZED : bool
        If True the session state has been initialized
        and the function does not need to run again.

    SK_USER : streamlit_passwordless.User or None, default None
        The current user of the application. If None a user has not signed
        in or registered yet.

    SK_USER_SIGN_IN : streamlit_passwordless.UserSignIn or None, default None
        Details from Bitwarden Passwordless about the user that signed in.
        If None a user has not signed in yet.

    SK_DB_USER : streamlit_passwordless.db.models.User or None, default None
        The database object of the current user. If None the user does not exist or
        has not been loaded from the database.

    SK_REGISTER_FORM_IS_VALID : bool, default False
        True if the input to the register form is valid and False otherwise.

    SK_REGISTER_FORM_VALIDATION_ERRORS : dict[str, str]
        A dictionary mapping of the form field name to its error message
        for the register form.

    SK_ROLES : dict[str, streamlit_passwordless.Role]
        The available roles for a user.

    SK_CUSTOM_ROLES : dict[str, streamlit_passwordless.CustomRole]
        The available custom roles for a user.

    SK_CREATE_USER_FORM_IS_VALID : bool
        True if the input to the create user form is valid and False otherwise.

    SK_CREATE_USER_FORM_VALIDATION_ERRORS : dict[str, str]
        A dictionary mapping of the form field name to its error message
        for the create user form.
    """

    if st.session_state.get(SK_SESSION_STATE_INITIALIZED, False):
        return

    st.session_state[SK_SESSION_STATE_INITIALIZED] = True

    none_keys = (SK_USER, SK_USER_SIGN_IN, SK_DB_USER)
    for key in none_keys:
        st.session_state[key] = None

    false_keys = (SK_CREATE_USER_FORM_IS_VALID, SK_REGISTER_FORM_IS_VALID)
    for key in false_keys:
        st.session_state[key] = False

    empty_dict_keys = (
        SK_REGISTER_FORM_VALIDATION_ERRORS,
        SK_CREATE_USER_FORM_VALIDATION_ERRORS,
        SK_ROLES,
        SK_CUSTOM_ROLES,
    )
    for key in empty_dict_keys:
        st.session_state[key] = {}


def get_current_user() -> User | None:
    r"""Get the current user from the session state.

    Returns
    -------
    streamlit_passwordless.User or None
        The user from the session state. None is returned if a user has not signed in yet.
    """

    return st.session_state.get(SK_USER)



r"""Functionality to check if a user is authorized to access content within an application."""

# Standard library
from functools import wraps
from typing import Callable, ParamSpec, TypeVar

# Third party
import streamlit as st
from streamlit.navigation.page import StreamlitPage

# Local
from streamlit_passwordless.components.config import SK_USER, SK_USER_SIGN_IN
from streamlit_passwordless.models import Role, User

P = ParamSpec('P')
R = TypeVar('R')


def authorized(
    user: User | None = None, role: Role | None = None, redirect: str | StreamlitPage | None = None
) -> Callable[[Callable[P, R | None]], Callable[P, R | None]]:
    r"""A decorator to check if a user is authorized to access content of specified `role`.

    If `user` is authorized the decorated function is executed and if not a redirect to
    another page is possible.

    Parameters
    ----------
    user : User or None, default None
        The user to authorize. If None the user to authorize is loaded from the
        session state using key :attr:`streamlit_passwordless.SK_USER`.

    role : streamlit_passwordless.Role or int or None, default None
        The role to authorize `user` against. If the rank of the role of `user` is
        greater than or equal to the rank of `role` the `user` is authorized. If an
        integer is supplied it is assumed to be the rank of the role to authorize the
        user against. If None (the default) the user is authorized regardless of its role.
        The user must always be authenticated to be authorized.

    redirect : str or streamlit.navigation.page.StreamlitPage or None, default None
        The Streamlit page to redirect `user` to if not authorized. If str it should be
        the path, relative to the file passed to the ``streamlit run`` command, to the
        Python file containing the page to redirect to. See :func:`streamlit.switch_page`
        for more info. If None, and `user` is not authorized, no redirect is performed and
        the decorated function is simply not executed.
    """

    def decorator(func: Callable[P, R | None]) -> Callable[P, R | None]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R | None:
            r"""Check if the user is authorized."""

            nonlocal user
            _user = st.session_state.get(SK_USER) if user is None else user

            if _user is None or not _user.is_authorized(role=role):
                if redirect is None:
                    return None
                st.switch_page(redirect)

            return func(*args, **kwargs)

        return wrapper

    return decorator


def sign_out(user: User | None) -> None:
    r"""Sign out a signed in user.

    Parameters
    ----------
    user : User or None, default None
        The user to sign out. If None the user to sign out is loaded from the
        session state using key :attr:`streamlit_passwordless.SK_USER`.
    """

    user = st.session_state.get(SK_USER) if user is None else user

    if user is not None:
        user.sign_in = None

    st.session_state[SK_USER_SIGN_IN] = None



r"""The database functionality of streamlit-passwordless."""

# Local
from . import models
from .cache import create_session_factory_cached, init
from .core import (
    URL,
    Session,
    SessionFactory,
    commit,
    create_db_url,
    create_session_factory,
)
from .crud.custom_role import create_custom_role, get_all_custom_roles, get_custom_roles
from .crud.email import get_email
from .crud.role import (
    create_default_roles,
    create_role,
    get_all_roles,
    get_role_by_name,
    get_role_by_role_id,
)
from .crud.user import (
    create_user,
    get_all_users,
    get_user_by_user_id,
    get_user_by_username,
)
from .crud.user_sign_in import create_user_sign_in
from .schemas.role import RoleCreate
from .schemas.user import UserCreate
from .schemas.user_sign_in import UserSignInCreate

# The Public API
__all__ = [
    'models',
    # cache
    'create_session_factory_cached',
    'init',
    # core
    'URL',
    'Session',
    'SessionFactory',
    'commit',
    'create_db_url',
    'create_session_factory',
    # custom_role
    'create_custom_role',
    'get_all_custom_roles',
    'get_custom_roles',
    # email
    'get_email',
    # role
    'create_default_roles',
    'create_role',
    'get_all_roles',
    'get_role_by_name',
    'get_role_by_role_id',
    # user
    'create_user',
    'get_all_users',
    'get_user_by_user_id',
    'get_user_by_username',
    # user_sign_in
    'create_user_sign_in',
    # schemas
    'RoleCreate',
    'UserCreate',
    'UserSignInCreate',
]



r"""The entry point of the initialization page."""

# Standard library
from pathlib import Path

# Third party
import streamlit as st

# Local
from streamlit_passwordless.app.config import (
    APP_HOME_PAGE_URL,
    APP_ISSUES_PAGE_URL,
    MAINTAINER_INFO,
    setup,
)
from streamlit_passwordless.app.controllers.init import controller

INIT_PATH = Path(__file__)

ABOUT = f"""Initialize the database of Streamlit Passwordless and create an admin account.

{MAINTAINER_INFO}
"""


def init_page() -> None:
    r"""Run the initialization page of Streamlit Passwordless."""

    st.set_page_config(
        page_title='Streamlit Passwordless - Initialize',
        page_icon=':sparkles:',
        layout='wide',
        menu_items={
            'Get Help': APP_HOME_PAGE_URL,
            'Report a bug': APP_ISSUES_PAGE_URL,
            'About': ABOUT,
        },
    )
    client, session_factory, cm = setup(create_database=True)

    with session_factory() as session:
        controller(cm=cm, db_session=session, client=client)


if __name__ == '__main__' or __name__ == '__page__':
    init_page()



r"""The views of the sign in page."""

# Third party
import streamlit as st


def title() -> None:
    r"""Render the title view of the sign in page."""

    st.title('Admin Console')
    st.subheader('Streamlit Passwordless')



r"""Helper functions and core components that can be used by other components."""

# Standard library
import logging
from enum import StrEnum
from functools import partial
from typing import Literal, Sequence, TypeAlias

# Third party
import streamlit as st
from streamlit.delta_generator import DeltaGenerator
from streamlit.navigation.page import StreamlitPage

# Local
from streamlit_passwordless import database as db
from streamlit_passwordless import exceptions, models
from streamlit_passwordless.bitwarden_passwordless.backend import BitwardenPasswordlessClient

from . import config

BannerContainer: TypeAlias = DeltaGenerator
ButtonType: TypeAlias = Literal['primary', 'secondary']
Redirectable: TypeAlias = StreamlitPage | str
logger = logging.getLogger(__name__)


class FormField(StrEnum):
    r"""The text input form fields defined in Streamlit Passwordless.

    Members
    -------
    USERNAME
        The username of a user.

    DISPLAYNAME
        The displayname of a user.

    AD_USERNAME
        The active directory username of a user.

    EMAIL
        An email address of a user.

    CREDENTIAL_NICKNAME
        The nickname for a passkey credential to register.
    """

    USERNAME = 'username'
    DISPLAYNAME = 'displayname'
    AD_USERNAME = 'ad_username'
    EMAIL = 'email'
    CREDENTIAL_NICKNAME = 'credential_nickname'


class BannerMessageType(StrEnum):
    r"""The banner message types defined in Streamlit Passwordless.

    Members
    -------
    INFO
        An info banner.

    SUCCESS
        A success banner.

    WARNING
        A warning banner.

    ERROR
        An error banner.
    """

    INFO = 'info'
    SUCCESS = 'success'
    WARNING = 'warning'
    ERROR = 'error'


def verify_sign_in(
    client: BitwardenPasswordlessClient, token: str
) -> tuple[models.UserSignIn | None, str]:
    r"""Verify the sign in token with the backend to complete the sign in process.

    Parameters
    ----------
    client : BitwardenPasswordlessClient
        The Bitwarden Passwordless client to use for interacting with
        the Bitwarden Passwordless API.

    token : str
        The token to verify.

    Returns
    -------
    user_sign_in : streamlit_passwordless.UserSignIn or None
        Details from Bitwarden Passwordless about the user that signed in.
        None is returned if an error occurred during the sign in process. `user_sign_in` is
        also stored in the session state with the key `config.SK_USER_SIGN_IN`.

    error_msg : str
        An error message about what failed during the sign in process.
        An empty string is returned if no errors occurred.
    """

    error_msg = ''
    user_sign_in: models.UserSignIn | None = None

    try:
        user_sign_in = client.verify_sign_in(token=token)
    except exceptions.SignInTokenVerificationError as e:
        error_msg = e.displayable_message
        logger.error(str(e))
    except exceptions.StreamlitPasswordlessError as e:
        error_msg = f'Error creating user sign in!\n{e.displayable_message}'
        logger.error(f'Error creating user sign in!\n{str(e)}')

    st.session_state[config.SK_USER_SIGN_IN] = user_sign_in

    return user_sign_in, error_msg


def delete_user_from_bitwarden_passwordless(
    client: BitwardenPasswordlessClient, user_id: str
) -> tuple[bool, str]:
    r"""Delete a user from Bitwarden Passwordless.

    Parameters
    ----------
    client : streamlit_passwordless.BitwardenPasswordlessClient
        The Bitwarden Passwordless client to use for interacting with
        the Bitwarden Passwordless API.

    user_id : str
        The unique ID of the user to delete.

    Returns
    -------
    success : bool
        True if the user was successfully deleted and False otherwise.

    error_msg : str
        An error message to display to the user if there was an issue with deleting
        the user. If no error occurred an empty string is returned.
    """

    try:
        client.delete_user(user_id=user_id)
    except exceptions.StreamlitPasswordlessError as e:
        logger.error(e.detailed_message)
        error_msg = e.displayable_message
        success = False
    else:
        error_msg = ''
        success = True

    return success, error_msg


def save_user_sign_in_to_database(
    session: db.Session, user_sign_in: models.UserSignIn
) -> tuple[bool, str]:
    r"""Save the user sign in entry to the database.

    Parameters
    ----------
    db_session : streamlit_passwordless.db.Session
        An active database session.

    user_sign_in : streamlit_passwordless.UserSignIn
        Data from Bitwarden Passwordless about the user that signed in.

    Returns
    -------
    success : bool
        True if the user sign in was successfully saved to the database and False otherwise.

    error_msg : str
        An error message to display to the user if there was an issue with saving the user sign
        in entry to the database. If no error occurred an empty string is returned.
    """

    user_sign_in_to_db = db.UserSignInCreate.model_validate(user_sign_in)

    try:
        db.create_user_sign_in(session=session, user_sign_in=user_sign_in_to_db, commit=True)
    except exceptions.DatabaseError as e:
        logger.error(e.detailed_message)
        error_msg = e.displayable_message
        success = False
    else:
        error_msg = ''
        success = True

    return success, error_msg


def get_user_from_database(
    session: db.Session, username: str, disabled: bool | None = None
) -> tuple[db.models.User | None, str]:
    r"""Get a user from the database.

    Updates the session state key `streamlit_passwordless.SK_DB_USER`
    with the user object from the database.

    Parameters
    ----------
    db_session : streamlit_passwordless.db.Session
        An active database session.

    username : str
        The username of the user to retrieve from the database.

    disabled : bool or None, default None
        Specify True to retrieve disabled users and False for enabled users.
        If None filtering by disabled or enabled user is omitted.

    Returns
    -------
    user : streamlit_passwordless.db.models.User or None
        The user matching `username`. None is returned if no user was found.

    error_msg : str
        An error message to display to the user if there was an issue with retrieving
        the user from the database. If no error occurred an empty string is returned.
    """

    try:
        user = db.get_user_by_username(session=session, username=username, disabled=disabled)
    except exceptions.DatabaseError as e:
        logger.error(e.detailed_message)
        error_msg = e.displayable_message
        user = None
    else:
        error_msg = ''

    st.session_state[config.SK_DB_USER] = user

    return user, error_msg


def create_user_in_database(
    session: db.Session,
    user: models.User,
    custom_roles: Sequence[db.models.CustomRole] | None = None,
    created_by_user_id: str | None = None,
) -> tuple[bool, str]:
    r"""Create a new user in the database.

    Parameters
    ----------
    session : streamlit_passwordless.db.Session
        An active database session.

    user : models.User
        The user to save to the database.

    custom_roles : Sequence[streamlit_passwordless.db.models.CustomRole] or None, default None
        The custom roles from the active database `session` to associate with the user.
        If provided these roles will take precedence over the custom roles defined on
        `user` and avoids a database lookup since the custom roles already exist in the `session`.

    created_by_user_id : str or None, default None
        The ID of the user that is creating the new user.

    Returns
    -------
    success : bool
        True if the user was successfully saved to the database and False otherwise.

    error_msg : str
        An error message to display to the user if there was an issue with creating the user
        in the database. If no error occurred an empty string is returned.
    """

    try:
        db.create_user(
            session=session,
            user=user,
            custom_roles=custom_roles,
            created_by_user_id=created_by_user_id,
            commit=True,
        )
    except exceptions.DatabaseCreateUserError as e:
        logger.error(e.detailed_message)
        error_msg = e.displayable_message
        success = False
    else:
        error_msg = ''
        success = True

    return success, error_msg


def update_user_in_database(session: db.Session, user: db.models.User) -> tuple[bool, str]:
    r"""Update an existing user in the database.

    Parameters
    ----------
    session : streamlit_passwordless.db.Session
        An active database session.

    user : models.User
        The user to update.

    Returns
    -------
    success : bool
        True if the user was successfully updated and False otherwise.

    error_msg : str
        An error message to display to the user if there was an issue with updating
        the user in the database. If no error occurred an empty string is returned.
    """

    try:
        db.commit(session, error_msg=f'Could not update {user.username}!')
    except exceptions.DatabaseError as e:
        logger.error(e.detailed_message)
        error_msg = e.displayable_message
        success = False
    else:
        error_msg = ''
        success = True

    return success, error_msg


def delete_user_in_database(session: db.Session, user: db.models.User) -> tuple[bool, str]:
    r"""Delete an existing user in the database.

    Parameters
    ----------
    session : streamlit_passwordless.db.Session
        An active database session.

    user : streamlit_passwordless.db.models.User
        The user to delete.

    Returns
    -------
    success : bool
        True if the user was successfully deleted and False otherwise.

    error_msg : str
        An error message to display to the user if there was an issue with deleting
        the user in the database. If no error occurred an empty string is returned.
    """

    try:
        session.delete(user)
        db.commit(session, error_msg=f'Could not delete {user.username}!')
    except exceptions.DatabaseError as e:
        logger.error(e.detailed_message)
        error_msg = e.displayable_message
        success = False
    else:
        error_msg = ''
        success = True

    return success, error_msg


def get_all_roles_from_database(session: db.Session) -> tuple[Sequence[db.models.Role], str]:
    r"""Get all roles from the database.

    Parameters
    ----------
    db_session : streamlit_passwordless.db.Session
        An active database session.

    Returns
    -------
    db_roles : Sequence[streamlit_passwordless.db.models.Role]
        The roles loaded from the database.

    error_msg : str
        An error message to display to the user if there was an issue with retrieving
        the roles from the database. If no error occurred an empty string is returned.
    """

    try:
        db_roles = db.get_all_roles(session=session, as_df=False)
    except exceptions.DatabaseError as e:
        logger.error(e.detailed_message)
        error_msg = e.displayable_message
        db_roles = []
    else:
        error_msg = ''

    return db_roles, error_msg


def get_all_custom_roles_from_database(
    session: db.Session,
) -> tuple[Sequence[db.models.CustomRole], str]:
    r"""Get all custom roles from the database.

    Parameters
    ----------
    db_session : streamlit_passwordless.db.Session
        An active database session.

    Returns
    -------
    db_custom_roles : Sequence[streamlit_passwordless.db.models.CustomRole]
        The custom roles from the database.

    error_msg : str
        An error message to display to the user if there was an issue with retrieving
        the custom roles from the database. If no error occurred an empty string is returned.
    """

    try:
        db_custom_roles = db.get_all_custom_roles(session=session, as_df=False)
    except exceptions.DatabaseError as e:
        logger.error(e.detailed_message)
        error_msg = e.displayable_message
        db_custom_roles = []
    else:
        error_msg = ''

    return db_custom_roles, error_msg


def get_email_from_database(
    session: db.Session, email: str, load_user: bool = False
) -> tuple[db.models.Email | None, str]:
    r"""Get an email from the database.

    Parameters
    ----------
    db_session : streamlit_passwordless.db.Session
        An active database session.

    email : str
        The email address to retrieve from the database.

    load_user : bool, default False
        If True the user that that the email belongs to will also be loaded from the database.

    Returns
    -------
    db_email : streamlit_passwordless.db.models.Email or None
        The email address matching `email` or None if an email was not found.

    error_msg : str
        An error message to display to the user if there was an issue with retrieving
        the email from the database. If no error occurred an empty string is returned.
    """

    try:
        db_email = db.get_email(session=session, email=email, load_user=load_user)
    except exceptions.DatabaseError as e:
        logger.error(e.detailed_message)
        error_msg = e.displayable_message
        db_email = None
    else:
        error_msg = ''

    return db_email, error_msg


def display_banner_message(
    message: str,
    message_type: BannerMessageType = BannerMessageType.SUCCESS,
    container: BannerContainer | None = None,
) -> None:
    r"""Display a message in a banner on a page.

    Parameters
    ----------
    message : str
        The message to display in the banner.

    message_type : BannerMessageType, default BannerMessageType.SUCCESS
        The type of message banner.

    container : streamlit_passwordless.BannerContainer or None, default None
        A container produced by :func:`streamlit.empty`, in which the banner will be displayed.
        Useful to make the banner appear at the desired location on a page. If None the banner
        will be displayed at the location of the page where the function is called.

    Raises
    ------
    streamlit_passwordless.StreamlitPasswordlessError
        If an invalid `message_type` is supplied.
    """

    banner_funcs = {
        BannerMessageType.INFO: partial(st.info, icon=config.ICON_INFO),
        BannerMessageType.SUCCESS: partial(st.success, icon=config.ICON_SUCCESS),
        BannerMessageType.WARNING: partial(st.warning, icon=config.ICON_WARNING),
        BannerMessageType.ERROR: partial(st.error, icon=config.ICON_ERROR),
    }
    func = banner_funcs.get(message_type)

    if func is None:
        valid_options = ', '.join(f'{v!r}' for v in BannerMessageType)
        error_msg = (
            f'Invalid value for display_banner_message({message_type=}). '
            f'Valid options are : {valid_options}'
        )
        logger.error(error_msg)
        raise exceptions.StreamlitPasswordlessError(message=error_msg, data=message_type)

    if container is None:
        func(message)
    else:
        with container:
            func(message)


def process_form_validation_errors(
    validation_errors: dict[str, str],
    banner_container_mapping: dict[str, BannerContainer],
    default_banner_container: BannerContainer | None = None,
) -> None:
    r"""Process form validation errors and display them in an error banner.

    Parameters
    ----------
    validation_errors : dict[str, str]
        A mapping of the field name to its error message.

    banner_container_mapping : dict[str, streamlit_passwordless.BannerContainer]
        A mapping of the field name to its banner container.

    default_banner_container : streamlit_passwordless.BannerContainer or None, default None
        The default banner container to use if a field does not have a matching entry in
        `banner_container_mapping`. If None the banner will be displayed at the location
        of the page where this function is called.
    """

    for field_name, error_msg in validation_errors.items():
        bc = banner_container_mapping.get(field_name, default_banner_container)
        display_banner_message(
            message=error_msg, message_type=BannerMessageType.ERROR, container=bc
        )



r"""The data models of streamlit-passwordless."""

# Standard library
import uuid
from datetime import datetime
from typing import Self

# Third party
from pydantic import AliasChoices
from pydantic import BaseModel as PydanticBaseModel
from pydantic import ConfigDict, Field, ValidationError, field_validator

# Local
from streamlit_passwordless.database.models import Role as DBRole

from . import exceptions


class BaseModel(PydanticBaseModel):
    r"""The BaseModel that all models inherit from."""

    model_config = ConfigDict(from_attributes=True, arbitrary_types_allowed=True)

    def __init__(self, **kwargs) -> None:
        try:
            super().__init__(**kwargs)
        except ValidationError as e:
            raise exceptions.StreamlitPasswordlessError(str(e)) from None


class BaseRole(BaseModel):
    r"""The base model for the models :class:`Role` and :class:`CustomRole`.

    :class:`BaseRole` should be subclassed and not used on its own.
    """

    model_config = ConfigDict(frozen=True)

    role_id: int | None = None
    name: str
    rank: int
    description: str | None = None

    def __eq__(self, other: object) -> bool:  # ==
        if isinstance(other, BaseRole):
            return self.rank == other.rank
        elif isinstance(other, int):
            return self.rank == other
        else:
            return NotImplemented

    def __nq__(self, other: object) -> bool:  # !=
        if isinstance(other, BaseRole):
            return self.rank != other.rank
        elif isinstance(other, int):
            return self.rank != other
        else:
            return NotImplemented

    def __lt__(self, other: object) -> bool:  # <
        if isinstance(other, BaseRole):
            return self.rank < other.rank
        elif isinstance(other, int):
            return self.rank < other
        else:
            return NotImplemented

    def __le__(self, other: object) -> bool:  # <=
        if isinstance(other, BaseRole):
            return self.rank <= other.rank
        elif isinstance(other, int):
            return self.rank <= other
        else:
            return NotImplemented

    def __gt__(self, other: object) -> bool:  # >
        if isinstance(other, BaseRole):
            return self.rank > other.rank
        elif isinstance(other, int):
            return self.rank > other
        else:
            return NotImplemented

    def __ge__(self, other: object) -> bool:  # >=
        if isinstance(other, BaseRole):
            return self.rank >= other.rank
        elif isinstance(other, int):
            return self.rank >= other
        else:
            return NotImplemented


class Role(BaseRole):
    r"""The role of a user.

    A :class:`User` is associated with a role to manage its privileges within an application.

    Parameters
    ----------
    role_id : int or None, default None
        The unique identifier of the role and the primary key in the database.
        If None the role is not persisted in the database.

    name : str
        The name of the role. Must be unique.

    rank : int
        The rank of the role. A role with a higher rank has more privileges. Used
        for comparing roles against one another. Two roles may have the same rank.

    description : str or None, default None
        A description of the role.
    """

    @classmethod
    def create_viewer(cls) -> Self:
        r"""Create the VIEWER role."""

        return cls.model_validate(DBRole.create_viewer())

    @classmethod
    def create_user(cls) -> Self:
        r"""Create the USER role, which is the default for a new user."""

        return cls.model_validate(DBRole.create_user())

    @classmethod
    def create_superuser(cls) -> Self:
        r"""Create the SUPERUSER role."""

        return cls.model_validate(DBRole.create_superuser())

    @classmethod
    def create_admin(cls) -> Self:
        r"""Create the ADMIN role."""

        return cls.model_validate(DBRole.create_admin())


ViewerRole = Role.create_viewer()
UserRole = Role.create_user()
SuperUserRole = Role.create_superuser()
AdminRole = Role.create_admin()


class CustomRole(BaseRole):
    r"""A custom role for a user.

    A :class:`User` may have none or multiple custom roles
    that are defined specifically for each application.

    Parameters
    ----------
    role_id : int or None, default None
        The unique identifier of the role and the primary key in the database.
        If None the role is not persisted in the database.

    name : str
        The name of the role. Must be unique.

    rank : int
        The rank of the role. A role with a higher rank has more privileges. Used
        for comparing roles against one another. Two roles may have the same rank.

    description : str or None, default None
        A description of the role.
    """


class Email(BaseModel):
    r"""An Email address of a user.

    Parameters
    ----------
    email_id : int or None, default None
        The unique identifier of the email and the primary key in the database.
        None is used when the email is not persisted in the database.

    user_id : str
        The unique ID of the user the email address belongs to.

    email : str
        An email address of a user. Must be unique across all users.

    rank : int
        The rank of the email, where 1 defines the primary email, 2 the secondary
        and 3 the tertiary etc ... A user can only have one email of each rank.

    verified : bool, default False
        True if the email address is verified and False otherwise.

    verified_at : datetime or None, default None
        The timestamp in UTC when the email address was verified by the user.

    disabled : bool, default False
        If the email address is disabled or not.

    disabled_at : datetime or None, default None
        The timestamp in UTC when the email address was disabled.
    """

    email_id: int | None = None
    user_id: str = ''
    email: str
    rank: int
    verified: bool = False
    verified_at: datetime | None = None
    disabled: bool = False
    disabled_at: datetime | None = None


class UserSignIn(BaseModel):
    r"""A model of a user sign in with a passkey credential.

    Parameters
    ----------
    user_sign_in_id : int or None, default None
        The unique ID of sign in entry. The primary key of the database table.
        None is used when the sign in entry is not persisted in the database.

    user_id : str
        The unique ID of the user that signed in to the application.

    sign_in_timestamp : datetime
        The timestamp in timezone UTC when the user signed in. Alias `timestamp`.

    success : bool
        True if the user was successfully signed in and False otherwise.

    origin : str
        The domain name or IP-address of the application the user signed in to.

    device : str
        The device the user signed in from. E.g. a web browser.

    country : str
        The country code of the country that the user signed in from. E.g. SE for Sweden.

    credential_nickname : str
        The nickname of the passkey that was used to sign in. Alias `nickname`.

    credential_id : str
        The ID of the passkey credential used to sign in.

    sign_in_type : str
        The type of sign in method. E.g. 'passkey_signin'. Alias `type`.

    rp_id : str or None
        The ID of the relaying party, which is the server that
        verifies the credentials during the sign in process.
    """

    user_sign_in_id: int | None = None
    user_id: str
    sign_in_timestamp: datetime = Field(
        validation_alias=AliasChoices('sign_in_timestamp', 'timestamp')
    )
    success: bool
    origin: str
    device: str
    country: str
    credential_nickname: str = Field(
        validation_alias=AliasChoices('credential_nickname', 'nickname')
    )
    credential_id: str
    sign_in_type: str = Field(validation_alias=AliasChoices('sign_in_type', 'type'))
    rp_id: str | None = None


class User(BaseModel):
    r"""A user within the streamlit-passwordless data model.

    Parameters
    ----------
    user_id : str, default ''
        The unique ID of the user which serves as the primary key in the database.
        If empty string (the default) it will be generated as a uuid.

    username : str
        The username of the user. It must be unique across all users.

    ad_username : str or None, default None
        The active directory username of the user if such exists.

    displayname : str or None, default None
        A descriptive name of the user that is easy to understand for a human.

    verified : bool, default False
        True if a user is verified and False otherwise. A user is verified when
        at least one verified email address is associated with the user.

    verified_at : datetime or None, default None
        The timestamp in UTC when the user was verified.

    disabled : bool, default False
        If False the user is enabled and if True the user is disabled.
        A disabled user is not able to register credentials or sign in.

    disabled_at : datetime or None, default None
        The timestamp in UTC when the user was disabled.

    role : streamlit_passwordless.Role, default streamlit_passwordless.UserRole
        The role of the user. The role is used for check if the user is authorized
        to access certain pages within an application.

    custom_roles : dict[str, streamlit_passwordless.CustomRole], default {}
        The custom roles of the user. The role name is mapped to the :class:`CustomRole` model.
        A user may have none or many custom roles.

    emails : list[streamlit_passwordless.Email], default []
        The email addresses associated with the user.

    sign_in : streamlit_passwordless.UserSignIn or None, default None
        Info about when the user signed in to the application.

    aliases : tuple[str, ...] or str or None, default None
        Additional ID:s that can be used to identify the user when signing in.
        A string with aliases separated by semicolon ";" can be used to supply multiple
        aliases if tuple is not used.
    """

    user_id: str = Field(default='', validate_default=True)
    username: str
    ad_username: str | None = None
    displayname: str | None = None
    verified: bool = False
    verified_at: datetime | None = None
    disabled: bool = False
    disabled_at: datetime | None = None
    role: Role = UserRole
    custom_roles: dict[str, CustomRole] = Field(default_factory=dict)
    emails: list[Email] = Field(default_factory=list)
    sign_in: UserSignIn | None = None
    aliases: tuple[str, ...] | str | None = Field(default=None, validate_default=True)

    def __hash__(self) -> int:
        return hash(self.user_id)

    @field_validator('user_id')
    def generate_user_id(cls, v: str) -> str:
        r"""Generate a user ID if not supplied."""

        return str(uuid.uuid4()) if not v else v

    @field_validator('aliases')
    def process_aliases(cls, aliases: tuple[str, ...] | str | None) -> tuple[str, ...] | None:
        r"""Convert multiple aliases in a string to a tuple by splitting on the semicolon ";"."""

        if isinstance(aliases, str):
            return tuple(v for a in aliases.split(';') if (v := a.strip()))
        else:
            return aliases

    @property
    def is_authenticated(self) -> bool:
        r"""Check if the user is authenticated with the application."""

        if (sign_in := self.sign_in) is None:
            return False
        else:
            return sign_in.user_id == self.user_id and sign_in.success

    def is_authorized(self, role: Role | int | None = None) -> bool:
        r"""Check if the user is authorized to access content of specified `role`.

        Parameters
        ----------
        role : streamlit_passwordless.Role or int or None, default None
            The role to authorize the user against. If the rank of the role of the user is
            greater than or equal to the rank of `role` the user is authorized. If an
            integer is supplied it is assumed to be the rank of the role to authorize the
            user against. If None (the default) the user is authorized regardless of its role.
            The user must always be authenticated to be authorized.

        Returns
        -------
        bool
            True if the user is authorized and False otherwise.
        """

        if role is None:
            return self.is_authenticated
        else:
            return self.role >= role if self.is_authenticated else False

    @property
    def email(self) -> str:
        r"""Get the primary email address of the user.

        An empty string is returned if the user does not have any email addresses.
        """

        try:
            return self.emails[0].email
        except IndexError:
            return ''



r"""The controller of the admin page."""

# Third party
import streamlit as st

# Local
import streamlit_passwordless.database as db
from streamlit_passwordless.app.logic.admin import load_users_and_roles_from_database
from streamlit_passwordless.app.views.admin import manage_users_view, sidebar, title
from streamlit_passwordless.bitwarden_passwordless import BitwardenPasswordlessClient
from streamlit_passwordless.models import User


def controller(client: BitwardenPasswordlessClient, db_session: db.Session, user: User) -> None:
    r"""Render the admin page.

    Parameters
    ----------
    client : streamlit_passwordless.BitwardenPasswordlessClient
        An instance of the Bitwarden Passwordless client to
        communicate with the backend API.

    db_session : streamlit_passwordless.db.Session
        An active session to the Streamlit Passwordless database.

    user : streamlit_passwordless.User
        The signed in admin user.
    """

    title()
    sidebar(user=user)
    users_tab, roles_tab, custom_roles_tab = st.tabs(('Users', 'Roles', 'Custom Roles'))

    df_users, roles, custom_roles = load_users_and_roles_from_database(db_session=db_session)

    with users_tab:
        manage_users_view(
            df=df_users,
            roles=roles,
            custom_roles=custom_roles,
            client=client,
            db_session=db_session,
        )
    with roles_tab:
        pass
    with custom_roles_tab:
        pass



r"""The Streamlit Passwordless administration web app."""

# Local
from .config import load_config, setup
from .pages.admin import admin_page
from .pages.init import init_page

# The Public API
__all__ = ['admin_page', 'init_page', 'load_config', 'setup']



r"""The core functionality of the schema models."""

# Standard library

# Third party
from pydantic import BaseModel as PydanticBaseModel
from pydantic import ConfigDict, ValidationError

# Local
from streamlit_passwordless import exceptions


class SchemaBaseModel(PydanticBaseModel):
    r"""The BaseModel that all schema models will inherit from."""

    model_config = ConfigDict(from_attributes=True)

    def __init__(self, **kwargs) -> None:
        try:
            super().__init__(**kwargs)
        except ValidationError as e:
            raise exceptions.StreamlitPasswordlessError(str(e)) from None



r"""The components of the admin app."""

# Standard library
from time import sleep
from typing import Sequence

# Third party
import pandas as pd
import streamlit as st

# Local
import streamlit_passwordless.database as db
from streamlit_passwordless.app.components import keys
from streamlit_passwordless.app.logic.admin import get_selectable_users_from_database
from streamlit_passwordless.components import create_user_form


@st.dialog(title='Create User')
def create_user_dialog_form(
    db_session: db.Session, on_success_exit_after_n_seconds: int | None = 1
) -> None:
    r"""Render the form for creating a new user in a dialog frame.

    Parameters
    ----------
    db_session : streamlit_passwordless.db.Session
        An active session to the Streamlit Passwordless database.

    on_success_exit_after_n_seconds : int, default 1
        The dialog form will close automatically after n number of seconds after a successful
        user creation. If None the dialog form will not close automatically.
    """

    created_user = create_user_form(
        db_session=db_session, with_ad_username=True, border=False, title=''
    )
    if created_user is not None:
        get_selectable_users_from_database(db_session=db_session)

        if on_success_exit_after_n_seconds is not None:
            sleep(on_success_exit_after_n_seconds)

        st.rerun(scope='app')


def user_state_radio_button(
    label: str = 'User State',
    default_disabled: bool = False,
    key: str = keys.ADMIN_USER_STATE_RADIO_BUTTON,
) -> bool:
    r"""Render the user state radio button.

    A user can be either enabled or disabled.

    Parameters
    ----------
    label : str, default 'User State'
        The label of the radio button.

    default_enabled : bool, default True
        True if the radio button will be set to 'enabled' on first render and
        False for 'disabled'.

    key : str, default streamlit_passwordless.app.keys.ADMIN_USER_STATE_RADIO_BUTTON
        The unique identifier of the component. Each component on a page must have a unique key.
    """

    return st.radio(
        label=label,
        options=(False, True),
        index=1 if default_disabled else 0,
        format_func=lambda x: 'Enabled' if x is False else 'Disabled',
        horizontal=True,
        key=key,
    )


def create_user_button(
    db_session: db.Session,
    label: str = 'Create User',
    disabled: bool = False,
    key=keys.ADMIN_CREATE_USER_BUTTON,
) -> bool:
    r"""Render the button for creating an new user.

    Parameters
    ----------
    db_session : streamlit_passwordless.db.Session
        An active session to the Streamlit Passwordless database.

    label : str, default 'Create User'
        The label of the button.

    disabled : bool, default False
        If True the component is disabled from user interaction.

    key : str, default streamlit_passwordless.app.keys.ADMIN_CREATE_USER_BUTTON
        The unique identifier of the component. Each component on a page must have a unique key.

    Returns
    -------
    bool
        True if the button was clicked and False otherwise.
    """

    return st.button(
        label=label,
        key=key,
        disabled=disabled,
        help='Create a new user in the database.',
        on_click=create_user_dialog_form,
        kwargs={'db_session': db_session},
    )


def refresh_users_button(
    db_session: db.Session,
    label: str = 'Refresh Users',
    disabled: bool = False,
    key=keys.ADMIN_REFRESH_USERS_BUTTON,
) -> bool:
    r"""Render button for refreshing the selectable users.

    Parameters
    ----------
    db_session : streamlit_passwordless.db.Session
        An active session to the Streamlit Passwordless database.

    label : str, default 'Refresh Users'
        The label of the button.

    disabled : bool, default False
        If True the component is disabled from user interaction.

    key : str, default streamlit_passwordless.app.keys.ADMIN_REFRESH_USERS_BUTTON
        The unique identifier of the component. Each component on a page must have a unique key.

    Returns
    -------
    bool
        True if the button was clicked and False otherwise.
    """

    return st.button(
        label=label,
        disabled=disabled,
        key=key,
        help='Load users from the database.',
        on_click=get_selectable_users_from_database,  # type: ignore
        kwargs={'db_session': db_session},
    )


def user_role_multiselect(
    roles: Sequence[db.models.Role],
    label: str = 'Roles',
    placeholder: str = 'Select the roles to filter by',
    key=keys.ADMIN_USER_ROLE_MULTISELECT,
) -> list[db.models.Role]:
    r"""Render the multi-selectbox of the roles to use for filtering the selectable users.

    Parameters
    ----------
    roles : Sequence[streamlit_passwordless.db.models.Role]
        The selectable roles.

    label : str, default 'Roles'
        The label of the multi-selectbox.

    placeholder : str, default 'Select the roles to filter by'
        The placeholder to display if no role is chosen.

    key : str, default streamlit_passwordless.app.keys.ADMIN_USER_ROLE_MULTISELECT
        The unique identifier of the component. Each component on a page must have a unique key.

    Returns
    -------
    list[streamlit_passwordless.db.models.Role]
        The selected roles.
    """

    return st.multiselect(
        label=label,
        options=roles,
        format_func=lambda r: r.name,
        key=key,
        placeholder=placeholder,
        help='Filter the users by roles.',
    )


def user_selectbox(
    df: pd.DataFrame, preselected_user: int | None = None, placeholder: str = 'Choose a user'
) -> str | None:
    r"""Render the selectbox for selecting the user to process.

    Parameters
    ----------
    df : pandas.DataFrame
        The DataFrame with the user info to provide for selecting a user. The index column
        should contain the user_id:s of the users. The columns of the DataFrame should be the
        searchable columns "username, "displayname", and "ad_username".

    preselected_user : int or None, default None
        The pre-selected user on first render of the component. 0 means the first user of `df`.
        If None no user is pre-selected and the value of `placeholder` is displayed instead.

    placeholder : str, default 'Choose a user'
        The placeholder to display if no user is chosen.

    key : str, default streamlit_passwordless.app.keys.ADMIN_USER_SELECTBOX
        The unique identifier of the component. Each component on a page must have a unique key.

    Returns
    -------
    str or None
        The selected user. None is returned if a user was not selected.
    """

    def format_user_display_text(user_id: str) -> str:
        r"""The text to display instead of the user_id in the selectbox."""

        s = df.loc[user_id, ['username', 'displayname', 'ad_username']]  # type: ignore

        return f'{s["username"]} | {s["displayname"]} | {s["ad_username"]}'

    nr_users = df.shape[0]
    label = (
        f'Select from {nr_users} user{"s" if nr_users > 0 else ""} '
        '(username | displayname | ad_username)'
    )
    return st.selectbox(
        label=label,
        options=df.index,
        index=preselected_user,  # type: ignore
        format_func=format_user_display_text,
        key=keys.ADMIN_USER_SELECTBOX,
        placeholder=placeholder,
        help='Search for a user by username, displayname or ad_username',
    )



r"""The components' library contains the web components of streamlit_passwordless."""

# Local
from .config import (
    ICON_ERROR,
    ICON_INFO,
    ICON_SUCCESS,
    ICON_WARNING,
    SK_CUSTOM_ROLES,
    SK_DB_USER,
    SK_REGISTER_FORM_IS_VALID,
    SK_ROLES,
    SK_SESSION_STATE_INITIALIZED,
    SK_USER,
    SK_USER_SIGN_IN,
    get_current_user,
    init_session_state,
)
from .core import (
    BannerContainer,
    BannerMessageType,
    Redirectable,
    display_banner_message,
    process_form_validation_errors,
)
from .create_user_form import create_user_form
from .delete_user_button import delete_user_button
from .register_form import bitwarden_register_form, bitwarden_register_form_existing_user
from .sign_in_form import bitwarden_sign_in_button, bitwarden_sign_in_form
from .sign_out import sign_out_button
from .update_user_form import update_user_form

# The Public API
__all__ = [
    # config
    'ICON_ERROR',
    'ICON_INFO',
    'ICON_SUCCESS',
    'ICON_WARNING',
    'SK_CUSTOM_ROLES',
    'SK_DB_USER',
    'SK_REGISTER_FORM_IS_VALID',
    'SK_ROLES',
    'SK_SESSION_STATE_INITIALIZED',
    'SK_USER',
    'SK_USER_SIGN_IN',
    'get_current_user',
    'init_session_state',
    # core
    'BannerContainer',
    'BannerMessageType',
    'Redirectable',
    'display_banner_message',
    'process_form_validation_errors',
    # create_user_form
    'create_user_form',
    # delete_user_button
    'delete_user_button',
    # register_form
    'bitwarden_register_form',
    'bitwarden_register_form_existing_user',
    # sign_in_form
    'bitwarden_sign_in_button',
    'bitwarden_sign_in_form',
    # sign_out
    'sign_out_button',
    # update_user_form
    'update_user_form',
]



r"""Database operations on the email table stp_email."""

# Standard library

# Third party
from sqlalchemy import select
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import selectinload

# Local
from streamlit_passwordless import exceptions
from streamlit_passwordless.database.core import Session
from streamlit_passwordless.database.models import Email


def get_email(session: Session, email: str, load_user: bool = False) -> Email | None:
    r"""Get an email from the database.

    Parameters
    ----------
    session : Session
        An active database session.

    email : str
        The email address to retrieve from the database.

    load_user : bool, default False
        If True the user that that the email belongs to will also be loaded from the database.

    Returns
    -------
    streamlit_passwordless.db.models.Email or None
        The email address matching `email` or None if an email was not found.

    Raises
    ------
    streamlit_passwordless.DatabaseError
        If an error occurs while loading the email address from the database.
    """

    query = select(Email).where(Email.email == email)
    if load_user:
        query = query.options(selectinload(Email.user))

    try:
        return session.scalars(query).one_or_none()
    except SQLAlchemyError as e:
        raise exceptions.DatabaseError(
            f'Error loading email "{email}" from database!', e=e
        ) from None



r"""The Bitwarden Passwordless package."""

# Local
from .backend import (
    BITWARDEN_PASSWORDLESS_API_URL,
    BitwardenPasswordlessClient,
    BitwardenRegisterConfig,
    PasskeyCredential,
)
from .cache import create_bitwarden_passwordless_client
from .frontend import register_button, sign_in_button

# The Public API

__all__ = [
    'BITWARDEN_PASSWORDLESS_API_URL',
    'BitwardenPasswordlessClient',
    'BitwardenRegisterConfig',
    'create_bitwarden_passwordless_client',
    'PasskeyCredential',
    'register_button',
    'sign_in_button',
]



r"""Functions for interacting with Bitwarden Passwordless that are cached by Streamlit."""

# Standard library
from datetime import timedelta

# Third party
import streamlit as st
from pydantic import AnyHttpUrl

# Local
from streamlit_passwordless.bitwarden_passwordless.backend import (
    BITWARDEN_PASSWORDLESS_API_URL,
    BitwardenPasswordlessClient,
    BitwardenRegisterConfig,
)


@st.cache_data(ttl=timedelta(days=7), show_spinner=False)
def create_bitwarden_passwordless_client(
    private_key: str,
    public_key: str,
    _url: AnyHttpUrl = BITWARDEN_PASSWORDLESS_API_URL,
    _register_config: BitwardenRegisterConfig | None = None,
) -> BitwardenPasswordlessClient:
    r"""Create the client to communicate with Bitwarden Passwordless.

    Parameters
    ----------
    public_key : str
        The public key of the Bitwarden Passwordless application.

    private_key : str
        The private key of the Bitwarden Passwordless application.

    _url : pydantic.AnyHttpUrl or str, default AnyHttpUrl('https://v4.passwordless.dev')
        The base url of the backend API of Bitwarden Passwordless. Specify this url
        if you are self-hosting Bitwarden Passwordless.

    _register_config : streamlit_passwordless.BitwardenRegisterConfig or None, default None
        The passkey configuration when registering a new user.
        If not specified the default configuration is used.

    Returns
    -------
    streamlit_passwordless.BitwardenPasswordlessClient
        An instance of the Bitwarden Passwordless client.
    """

    return BitwardenPasswordlessClient(
        public_key=public_key,
        private_key=private_key,
        url=_url,
        register_config=BitwardenRegisterConfig() if _register_config is None else _register_config,
    )



r"""Unit tests for the models' module."""

# Standard library
from copy import deepcopy
from datetime import datetime
from typing import Any, ClassVar, Sequence

# Third party
import pytest
from passwordless import VerifiedUser
from pydantic import ValidationError

# Local
from streamlit_passwordless import exceptions, models
from streamlit_passwordless.database import models as db_models

from .config import ModelData

# =============================================================================================
# Tests
# =============================================================================================


class TestBaseRole:
    r"""Tests for the model `BaseRole`."""

    user_role: ClassVar[models.BaseRole] = models.BaseRole(name='User', rank=1)
    admin_role: ClassVar[models.BaseRole] = models.BaseRole(name='Admin', rank=4)

    def test_init_with_defaults(self) -> None:
        r"""Test to initialize the model with required parameters only."""

        # Setup
        # ===========================================================
        input_data = {'name': 'USER', 'rank': 2}

        data_exp = deepcopy(input_data)
        data_exp['role_id'] = None
        data_exp['description'] = None

        # Exercise
        # ===========================================================
        role = models.Role.model_validate(input_data)

        # Verify
        # ===========================================================
        assert role.model_dump() == data_exp

        # Clean up - None
        # ===========================================================

    def test_init_with_all_parameters(self) -> None:
        r"""Test to supply values for all parameters."""

        # Setup
        # ===========================================================
        input_data = {'role_id': 1, 'name': 'USER', 'rank': 2, 'description': 'description'}

        # Exercise
        # ===========================================================
        role = models.Role.model_validate(input_data)

        # Verify
        # ===========================================================
        assert role.model_dump() == input_data

        # Clean up - None
        # ===========================================================

    @pytest.mark.parametrize(
        'role_left, role_right, exp_result',
        (
            pytest.param(user_role, user_role, True, id='True'),
            pytest.param(user_role, admin_role, False, id='False'),
            pytest.param(user_role, 1, True, id='int True'),
            pytest.param(user_role, 2, False, id='int False'),
        ),
    )
    def test___eq__(
        self, role_left: models.BaseRole, role_right: models.BaseRole | int, exp_result: bool
    ) -> None:
        r"""Test the `__eq__` (==) method."""

        # Setup - None
        # ===========================================================

        # Exercise
        # ===========================================================
        result = role_left == role_right

        # Verify
        # ===========================================================
        assert result == exp_result

        # Clean up - None
        # ===========================================================

    @pytest.mark.parametrize(
        'role_left, role_right, exp_result',
        (
            pytest.param(user_role, admin_role, True, id='True'),
            pytest.param(user_role, user_role, False, id='False'),
            pytest.param(user_role, 2, True, id='int True'),
            pytest.param(user_role, 1, False, id='int False'),
        ),
    )
    def test___nq__(
        self, role_left: models.BaseRole, role_right: models.BaseRole | int, exp_result: bool
    ) -> None:
        r"""Test the `__nq__` method (!=)."""

        # Setup - None
        # ===========================================================

        # Exercise
        # ===========================================================
        result = role_left != role_right

        # Verify
        # ===========================================================
        assert result == exp_result

        # Clean up - None
        # ===========================================================

    @pytest.mark.parametrize(
        'role_left, role_right, exp_result',
        (
            pytest.param(user_role, admin_role, True, id='True'),
            pytest.param(user_role, user_role, False, id='False'),
            pytest.param(user_role, 2, True, id='int True'),
            pytest.param(user_role, 1, False, id='int False'),
        ),
    )
    def test___lt__(
        self, role_left: models.BaseRole, role_right: models.BaseRole | int, exp_result: bool
    ) -> None:
        r"""Test the `__lt__` method (<)."""

        # Setup - None
        # ===========================================================

        # Exercise
        # ===========================================================
        result = role_left < role_right

        # Verify
        # ===========================================================
        assert result == exp_result

        # Clean up - None
        # ===========================================================

    @pytest.mark.parametrize(
        'role_left, role_right, exp_result',
        (
            pytest.param(user_role, admin_role, True, id='< True'),
            pytest.param(user_role, user_role, True, id='<= True'),
            pytest.param(admin_role, user_role, False, id='< False'),
            pytest.param(user_role, 2, True, id='< int True'),
            pytest.param(user_role, 1, True, id='<= int True'),
            pytest.param(user_role, 0, False, id='< int False'),
        ),
    )
    def test___le__(
        self, role_left: models.BaseRole, role_right: models.BaseRole | int, exp_result: bool
    ) -> None:
        r"""Test the `__le__` method (<=)."""

        # Setup - None
        # ===========================================================

        # Exercise
        # ===========================================================
        result = role_left <= role_right

        # Verify
        # ===========================================================
        assert result == exp_result

        # Clean up - None
        # ===========================================================

    @pytest.mark.parametrize(
        'role_left, role_right, exp_result',
        (
            pytest.param(admin_role, user_role, True, id='True'),
            pytest.param(user_role, admin_role, False, id='False'),
            pytest.param(user_role, 0, True, id='int True'),
            pytest.param(user_role, 1, False, id='int False'),
        ),
    )
    def test___gt__(
        self, role_left: models.BaseRole, role_right: models.BaseRole | int, exp_result: bool
    ) -> None:
        r"""Test the `__gt__` method (>)."""

        # Setup - None
        # ===========================================================

        # Exercise
        # ===========================================================
        result = role_left > role_right

        # Verify
        # ===========================================================
        assert result == exp_result

        # Clean up - None
        # ===========================================================

    @pytest.mark.parametrize(
        'role_left, role_right, exp_result',
        (
            pytest.param(admin_role, user_role, True, id='> True'),
            pytest.param(user_role, user_role, True, id='>= True'),
            pytest.param(user_role, admin_role, False, id='> False'),
            pytest.param(user_role, 0, True, id='> int True'),
            pytest.param(user_role, 1, True, id='>= int True'),
            pytest.param(user_role, 2, False, id='> int False'),
        ),
    )
    def test___ge__(
        self, role_left: models.BaseRole, role_right: models.BaseRole | int, exp_result: bool
    ) -> None:
        r"""Test the `__ge__` method (>=)."""

        # Setup - None
        # ===========================================================

        # Exercise
        # ===========================================================
        result = role_left >= role_right

        # Verify
        # ===========================================================
        assert result == exp_result

        # Clean up - None
        # ===========================================================


class TestRole:
    r"""Tests for the model `Role`."""

    def test_from_db_role(self, user_role: tuple[models.Role, db_models.Role, ModelData]) -> None:
        r"""Test to initialize the model from an instance of `database.models.Role`."""

        # Setup
        # ===========================================================
        _, db_role, role_data = user_role

        # Exercise
        # ===========================================================
        role = models.Role.model_validate(db_role)

        # Verify
        # ===========================================================
        assert role.model_dump() == role_data

        # Clean up - None
        # ===========================================================

    def test_create_viewer(
        self, viewer_role: tuple[models.Role, db_models.Role, ModelData]
    ) -> None:
        r"""Test to create a VIEWER role."""

        # Setup
        # ===========================================================
        _, _, role_data = viewer_role

        # Exercise
        # ===========================================================
        role = models.Role.create_viewer()

        # Verify
        # ===========================================================
        assert role.model_dump() == role_data

        # Clean up - None
        # ===========================================================

    def test_create_user(self, user_role: tuple[models.Role, db_models.Role, ModelData]) -> None:
        r"""Test to create a USER role."""

        # Setup
        # ===========================================================
        _, _, role_data = user_role

        # Exercise
        # ===========================================================
        role = models.Role.create_user()

        # Verify
        # ===========================================================
        assert role.model_dump() == role_data

        # Clean up - None
        # ===========================================================

    def test_create_superuser(
        self, superuser_role: tuple[models.Role, db_models.Role, ModelData]
    ) -> None:
        r"""Test to create a SUPERUSER role."""

        # Setup
        # ===========================================================
        _, _, role_data = superuser_role

        # Exercise
        # ===========================================================
        role = models.Role.create_superuser()

        # Verify
        # ===========================================================
        assert role.model_dump() == role_data

        # Clean up - None
        # ===========================================================

    def test_create_admin(self, admin_role: tuple[models.Role, db_models.Role, ModelData]) -> None:
        r"""Test to create an ADMIN role."""

        # Setup
        # ===========================================================
        _, _, role_data = admin_role

        # Exercise
        # ===========================================================
        role = models.Role.create_admin()

        # Verify
        # ===========================================================
        assert role.model_dump() == role_data

        # Clean up - None
        # ===========================================================

    def test_is_immutable(self) -> None:
        r"""Test that the model is immutable."""

        # Setup
        # ===========================================================
        role = models.Role(name='Immutable', rank=1)

        # Exercise & Verify
        # ===========================================================
        with pytest.raises(ValidationError):
            role.name = 'Mutable'  # type: ignore

        # Clean up - None
        # ===========================================================


class TestCustomRole:
    r"""Tests for the model `CustomRole`."""

    def test_from_db_custom_role(
        self, drummer_custom_role: tuple[models.CustomRole, db_models.CustomRole, ModelData]
    ) -> None:
        r"""Test to initialize the model from an instance of `database.models.CustomRole`."""

        # Setup
        # ===========================================================
        _, db_custom_role, custom_role_data = drummer_custom_role

        # Exercise
        # ===========================================================
        custom_role = models.CustomRole.model_validate(db_custom_role)

        # Verify
        # ===========================================================
        assert custom_role.model_dump() == custom_role_data

        # Clean up - None
        # ===========================================================


class TestEmail:
    r"""Tests for the model `Email`."""

    def test_init_with_defaults(self) -> None:
        r"""Test to initialize the model with required parameters only."""

        # Setup
        # ===========================================================
        input_data = {
            'user_id': 'user_id',
            'email': 'm.shadows@ax7.com',
            'rank': 1,
            'disabled': False,
        }

        data_exp = input_data.copy()
        data_exp['email_id'] = None
        data_exp['verified'] = False
        data_exp['verified_at'] = None
        data_exp['disabled_at'] = None

        # Exercise
        # ===========================================================
        user = models.Email.model_validate(input_data)

        # Verify
        # ===========================================================
        assert user.model_dump() == data_exp

        # Clean up - None
        # ===========================================================

    def test_init_with_all_parameters(self) -> None:
        r"""Test to supply values for all parameters."""

        # Setup
        # ===========================================================
        input_data = {
            'email_id': 1,
            'user_id': 'user_id',
            'email': 'test@email.com',
            'rank': 3,
            'verified': True,
            'verified_at': '2024-09-17T21:04:05',
            'disabled': True,
            'disabled_at': datetime(2024, 9, 9, 13, 37, 37),
        }

        data_exp = input_data.copy()
        data_exp['verified_at'] = datetime(2024, 9, 17, 21, 4, 5)

        # Exercise
        # ===========================================================
        user = models.Email.model_validate(input_data)

        # Verify
        # ===========================================================
        assert user.model_dump() == data_exp

        # Clean up - None
        # ===========================================================

    def test_from_db_email(
        self, user_1_email_primary: tuple[models.Email, db_models.Email, ModelData]
    ) -> None:
        r"""Test to initialize the model from an instance of `database.models.Email`."""

        # Setup
        # ===========================================================
        _, db_email, email_data = user_1_email_primary

        # Exercise
        # ===========================================================
        email = models.Email.model_validate(db_email)

        # Verify
        # ===========================================================
        assert email.model_dump() == email_data

        # Clean up - None
        # ===========================================================


class TestUserSignIn:
    r"""Tests for the model `UserSignIn`."""

    def test_init_with_default_values(self) -> None:
        r"""Test to initialize the model with required parameters only."""

        # Setup
        # ===========================================================
        input_data = {
            'user_id': 'user_id',
            'sign_in_timestamp': datetime(2024, 9, 18, 19, 20, 23),
            'success': True,
            'origin': 'origin',
            'device': 'device',
            'country': 'country',
            'credential_nickname': 'credential_nickname',
            'credential_id': 'credential_id',
            'sign_in_type': 'sign_in_type',
        }
        data_exp = input_data.copy()
        data_exp['user_sign_in_id'] = None
        data_exp['rp_id'] = None

        # Exercise
        # ===========================================================
        user_sign_in = models.UserSignIn.model_validate(input_data)

        # Verify
        # ===========================================================
        assert user_sign_in.model_dump() == data_exp

        # Clean up - None
        # ===========================================================

    def test_init_with_aliases(self) -> None:
        r"""Test to initialize the model with aliases.

        The following parameters are initialized with their aliases:
        - `sign_in_timestamp`   -> `timestamp`
        - `credential_nickname` -> `nickname`
        - `sign_in_type`        -> `type`
        """

        # Setup
        # ===========================================================
        input_data = {
            'user_sign_in_id': 1,
            'user_id': 'user_id',
            'timestamp': datetime(2024, 9, 18, 19, 20, 23),
            'success': True,
            'origin': 'origin',
            'device': 'device',
            'country': 'country',
            'nickname': 'credential_nickname',
            'credential_id': 'credential_id',
            'type': 'sign_in_type',
            'rp_id': 'rp_id',
        }
        data_exp = {
            'user_sign_in_id': 1,
            'user_id': 'user_id',
            'sign_in_timestamp': datetime(2024, 9, 18, 19, 20, 23),
            'success': True,
            'origin': 'origin',
            'device': 'device',
            'country': 'country',
            'credential_nickname': 'credential_nickname',
            'credential_id': 'credential_id',
            'sign_in_type': 'sign_in_type',
            'rp_id': 'rp_id',
        }

        # Exercise
        # ===========================================================
        user_sign_in = models.UserSignIn.model_validate(input_data)

        # Verify
        # ===========================================================
        assert user_sign_in.model_dump() == data_exp

        # Clean up - None
        # ===========================================================

    def test_from_db_user_sign_in(
        self, user_1_sign_in_unsuccessful: tuple[models.UserSignIn, db_models.UserSignIn, ModelData]
    ) -> None:
        r"""Test to initialize the model from an instance of `database.models.UserSignIn`."""

        # Setup
        # ===========================================================
        _, db_user_sign_in, user_sign_in_data = user_1_sign_in_unsuccessful

        # Exercise
        # ===========================================================
        user_sign_in = models.UserSignIn.model_validate(db_user_sign_in)

        # Verify
        # ===========================================================
        assert user_sign_in.model_dump() == user_sign_in_data

        # Clean up - None
        # ===========================================================

    def test_from_bitwarden_passwordless_verified_user(self) -> None:
        r"""Test to initialize the model from an instance of `passwordless.VerifiedUser`."""

        # Setup
        # ===========================================================
        verified_user = VerifiedUser(
            user_id='user_id',
            timestamp=datetime(2024, 9, 18, 19, 20, 23),
            success=False,
            origin='origin',
            device='device',
            country='country',
            nickname='credential_nickname',
            credential_id='credential_id',
            expires_at=datetime(2024, 9, 19, 19, 20, 23),
            token_id='token_id',
            type='sign_in_type',
            rp_id='rp_id',
        )
        data_exp = {
            'user_sign_in_id': None,
            'user_id': 'user_id',
            'sign_in_timestamp': datetime(2024, 9, 18, 19, 20, 23),
            'success': False,
            'origin': 'origin',
            'device': 'device',
            'country': 'country',
            'credential_nickname': 'credential_nickname',
            'credential_id': 'credential_id',
            'sign_in_type': 'sign_in_type',
            'rp_id': 'rp_id',
        }

        # Exercise
        # ===========================================================
        user_sign_in = models.UserSignIn.model_validate(verified_user)

        # Verify
        # ===========================================================
        assert user_sign_in.model_dump() == data_exp

        # Clean up - None
        # ===========================================================


class TestUser:
    r"""Tests for the model `User`."""

    def test_init_with_defaults(
        self, mocked_user_id: str, user_role: tuple[models.Role, db_models.Role, ModelData]
    ) -> None:
        r"""Test to initialize the model with required parameters only.

        By default a user is created with the `models.UserNameRole.USER` role.
        """

        # Setup
        # ===========================================================
        _, _, role_data = user_role

        username = 'm.shadows'
        data_exp: dict[str, Any] = {
            'user_id': mocked_user_id,
            'username': username,
            'ad_username': None,
            'displayname': None,
            'verified': False,
            'verified_at': None,
            'disabled': False,
            'disabled_at': None,
            'role': role_data,
            'custom_roles': {},
            'emails': [],
            'sign_in': None,
            'aliases': None,
        }

        # Exercise
        # ===========================================================
        user = models.User(username=username)

        # Verify
        # ===========================================================
        assert user.model_dump() == data_exp

        # Clean up - None
        # ===========================================================

    def test_init_with_all_parameters(
        self,
        user_1_email_primary: tuple[models.Email, db_models.Email, ModelData],
        user_1_email_secondary: tuple[models.Email, db_models.Email, ModelData],
        user_1_sign_in_successful: tuple[models.UserSignIn, db_models.UserSignIn, ModelData],
    ) -> None:
        r"""Test supply values for all parameters of the model."""

        # Setup
        # ===========================================================
        _, _, email_primary_data = user_1_email_primary
        _, _, email_secondary_data = user_1_email_secondary
        _, _, sign_in_data = user_1_sign_in_successful

        input_data = {
            'user_id': 'user_id',
            'username': 'username',
            'ad_username': 'ad_username',
            'displayname': 'displayname',
            'verified': False,
            'verified_at': '2024-09-17T20:48:16',
            'disabled': True,
            'disabled_at': '2024-09-18T21:48:16',
            'role': {'role_id': 1, 'name': 'role', 'rank': 1, 'description': None},
            'custom_roles': {
                'custom_role_1': {
                    'role_id': 1,
                    'name': 'custom_role',
                    'rank': 1,
                    'description': 'description',
                },
            },
            'emails': [email_primary_data, email_secondary_data],
            'sign_in': sign_in_data,
            'aliases': 'Matt;Shadows',
        }

        data_exp = deepcopy(input_data)
        data_exp['verified_at'] = datetime(2024, 9, 17, 20, 48, 16)
        data_exp['disabled_at'] = datetime(2024, 9, 18, 21, 48, 16)
        data_exp['aliases'] = ('Matt', 'Shadows')

        # Exercise
        # ===========================================================
        user = models.User.model_validate(input_data)

        # Verify
        # ===========================================================
        assert user.model_dump() == data_exp

        # Clean up - None
        # ===========================================================

    def test_from_db_user(self, user_1: tuple[models.User, db_models.User, ModelData]) -> None:
        r"""Test to initialize the model from `database.models.User`."""

        # Setup
        # ===========================================================
        _, db_user, user_data = user_1

        data_exp = deepcopy(user_data)
        data_exp['emails'] = []

        # Exercise
        # ===========================================================
        user = models.User.model_validate(db_user)

        # Verify
        # ===========================================================
        assert user.model_dump() == data_exp

        # Clean up - None
        # ===========================================================

    def test_is_authenticated_user_sign_in_none(
        self, user_1: tuple[models.User, db_models.User, ModelData]
    ) -> None:
        r"""Test that a user that is not signed in is not listed as authenticated."""

        # Setup
        # ===========================================================
        user, _, _ = user_1

        # Exercise & Verify
        # ===========================================================
        assert user.is_authenticated is False, 'user.is_authenticated is True!'
        assert user.sign_in is None, 'user.sign_in is not None!'

        # Clean up - None
        # ===========================================================

    def test_is_authenticated_user_sign_in_success_true(
        self,
        user_1_with_2_emails_and_successful_signin: tuple[models.User, db_models.User, ModelData],
    ) -> None:
        r"""Test that a user that successfully signed in with a passkey is authenticated."""

        # Setup
        # ===========================================================
        user, _, _ = user_1_with_2_emails_and_successful_signin

        # Exercise & Verify
        # ===========================================================
        assert user.is_authenticated is True, 'user.is_authenticated is False!'
        assert getattr(user.sign_in, 'success') is True, 'user.sign_in.success is False!'

        # Clean up - None
        # ===========================================================

    def test_is_authenticated_user_sign_in_success_false(
        self, user_1_with_unsuccessful_signin: tuple[models.User, db_models.User, ModelData]
    ) -> None:
        r"""Test that a user with a failed passkey sign in is not authenticated."""

        # Setup
        # ===========================================================
        user, _, _ = user_1_with_unsuccessful_signin

        # Exercise & Verify
        # ===========================================================
        assert user.is_authenticated is False, 'user.is_authenticated is True!'
        assert getattr(user.sign_in, 'success') is False, 'user.sign_in.success is True!'

        # Clean up - None
        # ===========================================================

    def test_is_authenticated_user_sign_in_user_id_not_equal_to_user_user_id(
        self, user_1_sign_in_successful: tuple[models.UserSignIn, db_models.UserSignIn, ModelData]
    ) -> None:
        r"""Test comparing `user.user_id` to `user.sign_in.user_id` when checking authentication.

        If `user.user_id` differs from `user.sign_in.user_id` the user should not be
        authenticated. This safe guards against if a `models.UserSignIn` instance for one
        user is accidentally assigned to another user.
        """

        # Setup
        # ===========================================================
        user_sign_in, _, _ = user_1_sign_in_successful
        user = models.User(user_id='user_id_not_equal_user_sign_in_user_id', username='username')
        user.sign_in = user_sign_in

        # Exercise & Verify
        # ===========================================================
        assert user.is_authenticated is False, 'user.is_authenticated is True!'
        assert getattr(user.sign_in, 'success') is True, 'user.sign_in.success is False!'

        # Clean up - None
        # ===========================================================

    def test_get_primary_email(self) -> None:
        r"""Test to get the primary email of the user."""

        # Setup
        # ===========================================================
        exp_email = 'test@example.com'
        user = models.User(
            username='username',
            emails=[
                models.Email(email=exp_email, rank=1),
                models.Email(email='test2@example.com', rank=2),
            ],
        )

        # Exercise & Verify
        # ===========================================================
        assert user.email == exp_email

        # Clean up - None
        # ===========================================================

    def test_get_primary_email_user_has_no_email_addresses(self) -> None:
        r"""Test to get the primary email of the user.

        If the user has no email addresses an empty string should be returned.
        """

        # Setup
        # ===========================================================
        user = models.User(username='username')

        # Exercise & Verify
        # ===========================================================
        assert user.email == ''

        # Clean up - None
        # ===========================================================

    @pytest.mark.parametrize(
        'aliases, aliases_exp',
        (
            pytest.param(None, None, id='None'),
            pytest.param(('Matt', 'Shadows'), ('Matt', 'Shadows'), id="('Matt', 'Shadows')"),
            pytest.param(['Matt', 'Shadows'], ('Matt', 'Shadows'), id="['Matt', 'Shadows']"),
            pytest.param('Matt;Shadows', ('Matt', 'Shadows'), id='Matt;Shadows;'),
            pytest.param('Matt', ('Matt',), id='str'),
            pytest.param('Matt;', ('Matt',), id='Matt;'),
        ),
    )
    def test_aliases(
        self, aliases: str | Sequence[str], aliases_exp: tuple[str, ...] | None
    ) -> None:
        r"""Test to supply different types of aliases."""

        # Setup
        # ===========================================================
        username = 'm.shadows'

        # Exercise
        # ===========================================================
        user = models.User(username=username, aliases=aliases)

        # Verify
        # ===========================================================
        assert user.aliases == aliases_exp

        # Clean up - None
        # ===========================================================

    @pytest.mark.raises
    def test_invalid_username(self) -> None:
        r"""Test supply an invalid username.

        `exceptions.StreamlitPasswordlessError` is expected to be raised.
        """

        # Setup - None
        # ===========================================================

        # Exercise
        # ===========================================================
        with pytest.raises(exceptions.StreamlitPasswordlessError) as exc_info:
            models.User(username=None)

        # Verify
        # ===========================================================
        error_msg = exc_info.exconly()
        print(error_msg)

        assert 'username' in error_msg, 'username not in error message!'

        # Clean up - None
        # ===========================================================


class TestUserIsAuthorized:
    r"""Tests for the method `User.is_authorized`."""

    @pytest.mark.parametrize(
        'role_as_int', (pytest.param(False, id='Role'), pytest.param(True, id='int'))
    )
    def test_authenticated_and_authorized_superuser_against_user(
        self,
        role_as_int: bool,
        user_1_with_2_emails_and_successful_signin: tuple[models.User, db_models.User, ModelData],
        user_role: tuple[models.Role, db_models.Role, ModelData],
    ) -> None:
        r"""Test that an authenticated superuser is authorized against a user role."""

        # Setup
        # ===========================================================
        user, _, _ = user_1_with_2_emails_and_successful_signin
        authorized_role, _, _ = user_role
        role: models.Role | int = authorized_role.rank if role_as_int else authorized_role

        # Exercise & Verify
        # ===========================================================
        assert user.is_authorized(role=role) is True

        # Clean up - None
        # ===========================================================

    @pytest.mark.parametrize(
        'role_as_int', (pytest.param(False, id='Role'), pytest.param(True, id='int'))
    )
    def test_authenticated_and_authorized_superuser_against_superuser(
        self,
        role_as_int: bool,
        user_1_with_2_emails_and_successful_signin: tuple[models.User, db_models.User, ModelData],
        superuser_role: tuple[models.Role, db_models.Role, ModelData],
    ) -> None:
        r"""Test that an authenticated superuser is authorized against a superuser role."""

        # Setup
        # ===========================================================
        user, _, _ = user_1_with_2_emails_and_successful_signin
        authorized_role, _, _ = superuser_role
        role: models.Role | int = authorized_role.rank if role_as_int else authorized_role

        # Exercise & Verify
        # ===========================================================
        assert user.is_authorized(role=role) is True

        # Clean up - None
        # ===========================================================

    @pytest.mark.parametrize(
        'role_as_int', (pytest.param(False, id='Role'), pytest.param(True, id='int'))
    )
    def test_authenticated_and_not_authorized_superuser_against_admin(
        self,
        role_as_int: bool,
        user_1_with_2_emails_and_successful_signin: tuple[models.User, db_models.User, ModelData],
        admin_role: tuple[models.Role, db_models.Role, ModelData],
    ) -> None:
        r"""Test that an authenticated superuser is not authorized against an admin role."""

        # Setup
        # ===========================================================
        user, _, _ = user_1_with_2_emails_and_successful_signin
        authorized_role, _, _ = admin_role
        role: models.Role | int = authorized_role.rank if role_as_int else authorized_role

        # Exercise & Verify
        # ===========================================================
        assert user.is_authorized(role=role) is False

        # Clean up - None
        # ===========================================================

    @pytest.mark.parametrize(
        'role_as_int', (pytest.param(False, id='Role'), pytest.param(True, id='int'))
    )
    def test_not_authenticated_and_authorized_superuser_against_user(
        self,
        role_as_int: bool,
        user_1_with_unsuccessful_signin: tuple[models.User, db_models.User, ModelData],
        user_role: tuple[models.Role, db_models.Role, ModelData],
    ) -> None:
        r"""Test that a non-authenticated superuser is not authorized against a user role."""

        # Setup
        # ===========================================================
        user, _, _ = user_1_with_unsuccessful_signin
        authorized_role, _, _ = user_role
        role: models.Role | int = authorized_role.rank if role_as_int else authorized_role

        # Exercise & Verify
        # ===========================================================
        assert user.is_authorized(role=role) is False

        # Clean up - None
        # ===========================================================

    @pytest.mark.parametrize(
        'role_as_int', (pytest.param(False, id='Role'), pytest.param(True, id='int'))
    )
    def test_not_authenticated_and_authorized_superuser_against_superuser(
        self,
        role_as_int: bool,
        user_1_with_unsuccessful_signin: tuple[models.User, db_models.User, ModelData],
        superuser_role: tuple[models.Role, db_models.Role, ModelData],
    ) -> None:
        r"""Test that a non-authenticated superuser is not authorized against a superuser role."""

        # Setup
        # ===========================================================
        user, _, _ = user_1_with_unsuccessful_signin
        authorized_role, _, _ = superuser_role
        role: models.Role | int = authorized_role.rank if role_as_int else authorized_role

        # Exercise & Verify
        # ===========================================================
        assert user.is_authorized(role=role) is False

        # Clean up - None
        # ===========================================================

    @pytest.mark.parametrize(
        'role_as_int', (pytest.param(False, id='Role'), pytest.param(True, id='int'))
    )
    def test_not_authenticated_and_not_authorized_superuser_against_admin(
        self,
        role_as_int: bool,
        user_1_with_unsuccessful_signin: tuple[models.User, db_models.User, ModelData],
        admin_role: tuple[models.Role, db_models.Role, ModelData],
    ) -> None:
        r"""Test that a non-authenticated superuser is not authorized against an admin role."""

        # Setup
        # ===========================================================
        user, _, _ = user_1_with_unsuccessful_signin
        authorized_role, _, _ = admin_role
        role: models.Role | int = authorized_role.rank if role_as_int else authorized_role

        # Exercise & Verify
        # ===========================================================
        assert user.is_authorized(role=role) is False

        # Clean up - None
        # ===========================================================

    def test_authenticated_and_role_is_none(
        self,
        user_1_with_2_emails_and_successful_signin: tuple[models.User, db_models.User, ModelData],
    ) -> None:
        r"""Test that an authenticated user is authorized when `role` parameter is None."""

        # Setup
        # ===========================================================
        user, _, _ = user_1_with_2_emails_and_successful_signin

        # Exercise & Verify
        # ===========================================================
        assert user.is_authorized(role=None) is True

        # Clean up - None
        # ===========================================================

    def test_not_authenticated_and_role_is_none(
        self, user_1_with_unsuccessful_signin: tuple[models.User, db_models.User, ModelData]
    ) -> None:
        r"""Test that a non-authenticated user is not authorized when `role` parameter is None."""

        # Setup
        # ===========================================================
        user, _, _ = user_1_with_unsuccessful_signin

        # Exercise & Verify
        # ===========================================================
        assert user.is_authorized(role=None) is False

        # Clean up - None
        # ===========================================================



r"""The application logic of the admin app."""

# Standard library
from typing import Sequence

# Third party
import pandas as pd
import streamlit as st

# Local
import streamlit_passwordless.database as db
from streamlit_passwordless.app.session_state import (
    SK_DB_CUSTOM_ROLES,
    SK_DB_ROLES,
    SK_SELECTABLE_USERS_DATAFRAME,
)


def get_selectable_users_from_database(db_session: db.Session) -> pd.DataFrame:
    r"""Load the selectable users from the database.

    Updates the session state key
    :attr:`streamlit_passwordless.app.components.keys.SK_SELECTABLE_USERS_DATAFRAME`
    with the function result.

    Parameters
    ----------
    db_session : streamlit_passwordless.db.Session
        An active session to the Streamlit Passwordless database.

    Returns
    -------
    pandas.DataFrame
        The selectable users.
    """

    df = db.get_all_users(session=db_session, as_df=True)
    st.session_state[SK_SELECTABLE_USERS_DATAFRAME] = df

    return df


def get_all_roles_from_database(db_session: db.Session) -> Sequence[db.models.Role]:
    r"""Load the user roles from the database.

    Updates the session state key
    :attr:`streamlit_passwordless.app.components.keys.SK_DB_ROLES`
    with the function result.

    Parameters
    ----------
    db_session : streamlit_passwordless.db.Session
        An active session to the Streamlit Passwordless database.

    Returns
    -------
    roles : Sequence[streamlit_passwordless.db.models.Role]
        The user roles.
    """

    roles = db.get_all_roles(session=db_session, as_df=False)
    st.session_state[SK_DB_ROLES] = roles

    return roles


def get_all_custom_roles_from_database(db_session: db.Session) -> Sequence[db.models.CustomRole]:
    r"""Load the user custom roles from the database.

    Updates the session state key
    :attr:`streamlit_passwordless.app.components.keys.SK_DB_CUSTOM_ROLES`
    with the function result.

    Parameters
    ----------
    db_session : streamlit_passwordless.db.Session
        An active session to the Streamlit Passwordless database.

    Returns
    -------
    custom_roles : Sequence[streamlit_passwordless.db.models.Role]
        The user custom roles.
    """

    custom_roles = db.get_all_custom_roles(session=db_session, as_df=False)
    st.session_state[SK_DB_CUSTOM_ROLES] = custom_roles

    return custom_roles


def load_users_and_roles_from_database(
    db_session: db.Session,
) -> tuple[pd.DataFrame, Sequence[db.models.Role], Sequence[db.models.CustomRole]]:
    r"""Load the selectable users, roles and custom roles from the database.

    Parameters
    ----------
    db_session : streamlit_passwordless.db.Session
        An active session to the Streamlit Passwordless database.

    Returns
    -------
    df_users : pandas.DataFrame
        The selectable users.

    roles : Sequence[streamlit_passwordless.db.models.Role]
        The user roles.

    custom_roles : Sequence[streamlit_passwordless.db.models.CustomRole]
        The user custom roles.
    """

    if st.session_state.get(SK_SELECTABLE_USERS_DATAFRAME) is None:
        get_selectable_users_from_database(db_session=db_session)
    df_users = st.session_state.get(SK_SELECTABLE_USERS_DATAFRAME, pd.DataFrame())

    if st.session_state.get(SK_DB_ROLES) is None:
        get_all_roles_from_database(db_session=db_session)
    roles = st.session_state.get(SK_DB_ROLES, [])

    if st.session_state.get(SK_DB_CUSTOM_ROLES) is None:
        get_all_custom_roles_from_database(db_session=db_session)
    custom_roles = st.session_state.get(SK_DB_CUSTOM_ROLES, [])

    return df_users, roles, custom_roles


def filter_selectable_users(
    df: pd.DataFrame, disabled: bool | None, roles: set[int]
) -> pd.DataFrame:
    r"""Filter the selectable users.

    Parameters
    ----------
    df : pandas.DataFrame
        The selectable users.

    disabled : bool or None
        Filter by disabled (True) or enabled (False) users.
        If None filtering by disabled/enabled is omitted.

    roles : set[int]
        The role_id:s of the user roles to filter by.

    Returns
    -------
    pandas.DataFrame
        The selectable users after filtering by disabled/enabled and roles.
    """

    if disabled is None:
        if roles:
            return df.loc[df['role_id'].isin(roles), :]
        else:
            return df
    else:
        disabled_filter = df['disabled'].eq(disabled)
        if roles:
            roles_filter = df['role_id'].isin(roles)
            return df.loc[disabled_filter & roles_filter, :]
        else:
            return df.loc[disabled_filter, :]
